part of geoclip;

class Boolean {
  static final bool _DEBUG = false;

  static final double _EPSILON = 2.2204460492503131e-016;

  static const int LEFT = 0;
  static const int RIGHT = 1;

  static const int ABOVE = 0;
  static const int BELOW = 1;

  // clip applied to Subjects.
  static const int CLIP = 0;
  // Subjects
  static const int SUBJ = 1;

  /**
   * Return the Intersection of [p1] and [p2].
   */
  static Polygon intersection(Polygon p1, Polygon p2, String classType) {
    return clip(OperationType.INT, p1, p2, classType);
  }

  static Polygon intersectionComplex(Polygon p1, Polygon p2) {
    return clip(OperationType.INT, p1, p2, "ComplexPolygon");
  }

  /**
   * Return the Union or Add of [p1] and [p2].
   */
  static Polygon add(Polygon p1, Polygon p2, String classType) {
    return clip(OperationType.UNION, p1, p2, classType);
  }

  static Polygon addComplex(Polygon p1, Polygon p2) {
    return clip(OperationType.UNION, p1, p2, "ComplexPolygon");
  }
  
  /**
   * Return the Exclusive-Or of [p1] and [p2].
   */
  static Polygon xor(Polygon p1, Polygon p2, String classType) {
    return clip(OperationType.XOR, p1, p2, classType);
  }

  static Polygon xorComplex(Polygon p1, Polygon p2) {
    return clip(OperationType.XOR, p1, p2, "ComplexPolygon");
  }

  /**
   * Return the Difference or Subtraction of [p1] and [p2].
   */
  static Polygon subtract(Polygon p1, Polygon p2, String classType) {
    return clip(OperationType.DIFF, p1, p2, classType);
  }

  /**
   * subtract [p2] from [p1]
   */
  static Polygon subtractComplex(Polygon p1, Polygon p2) {
    return clip(OperationType.DIFF, p1, p2, "ComplexPolygon");
  }
  
  static bool EQ(double a, double b) {
    // TODO this isn't numerically robust.
    return ((a - b).abs() <= _EPSILON);
  }

  static int PREV_INDEX( int i, int n) {
     return ((i - 1 + n) % n);
  }
  
  static int NEXT_INDEX(int i, int n) {
     return ((i + 1) % n);
  }

  static bool OPTIMAL(Polygon p, int i) {
    GPoint prev = p.getPoint(PREV_INDEX(i, p.numberOfPoints));
    GPoint next = p.getPoint(NEXT_INDEX(i, p.numberOfPoints));
    GPoint pp = p.getPoint(i);
    return (prev != pp) || (next != pp); 
  }

  
  static _ItNode _add_intersection(_ItNode it_node, _EdgeNode edge0, _EdgeNode  edge1, double x, double y) {
    if (it_node == null) {
      // Append a new node to the tail of the list
      it_node = new _ItNode( edge0, edge1, x, y, null );
    }
    else {
      if (it_node.point.y > y) {
        // Insert a new node mid-list
        _ItNode existing_node = it_node;
        it_node = new _ItNode( edge0, edge1, x, y, existing_node );
      }
      else {
        // Head further down the list
        it_node.next = _add_intersection( it_node.next, edge0, edge1, x, y);
      }
    }
    return it_node;
  }

  static _StNode _add_st_edge(_StNode st, _ItNodeTable it, _EdgeNode edge, double dy) {
    if (st == null) {
      // Append edge onto the tail end of the ST
      st = new _StNode(edge, null);
    }
    else {
      double den= (st.xt - st.xb) - (edge.xt - edge.xb);
       
      // If new edge and ST edge don't cross
      if( (edge.xt >= st.xt) || (edge.dx == st.dx) || (den.abs() <= _EPSILON)) {
        // No intersection - insert edge here (before the ST edge)
        _StNode existing_node = st;
        st = new _StNode(edge, existing_node);
      }
      else {
        // Compute intersection between new edge and ST edge
        double r= (edge.xb - st.xb) / den;
        double x= st.xb + r * (st.xt - st.xb);
        double y= r * dy;
        
        // Insert the edge pointers and the intersection point in the IT
        it.top_node = _add_intersection(it.top_node, st.edge, edge, x, y);
        
        // Head further into the ST
        st.prev = _add_st_edge(st.prev, it, edge, dy);
     }
    }
    return st;
  }

  static void _add_edge_to_aet(_AetTree aet, _EdgeNode edge) {
    if ( aet.top_node == null ) {
      // Append edge onto the tail end of the AET
      aet.top_node = edge;
      edge.prev = null;
      edge.next= null;
    }
    else {
      _EdgeNode current_edge = aet.top_node;
      _EdgeNode prev = null;
      bool done = false;

      while(!done) {
        // Do primary sort on the xb field
        if (edge.xb < current_edge.xb) {
           // Insert edge here (before the AET edge)
           edge.prev= prev;
           edge.next= current_edge;
           current_edge.prev = edge;
           if( prev == null ) {
              aet.top_node = edge;
           }
           else {
              prev.next = edge;
           }
           done = true;
        }
        else if (edge.xb == current_edge.xb) {
          // Do secondary sort on the dx field
          if (edge.dx < current_edge.dx) {
             // Insert edge here (before the AET edge)
             edge.prev= prev;
             edge.next= current_edge;
             current_edge.prev = edge;
             if( prev == null) {
               aet.top_node = edge;
             }
             else {
               prev.next = edge;
             }
             done = true;
          }
          else {
             // Head further into the AET
             prev = current_edge;
             if (current_edge.next == null) {
               current_edge.next = edge;
               edge.prev = current_edge;
               edge.next = null;
               done = true;
             }
             else {
               current_edge = current_edge.next;
             }
          }
        }
        else {
          // Head further into the AET
          prev = current_edge;
          if (current_edge.next == null) {
            current_edge.next = edge;
            edge.prev = current_edge;
            edge.next = null;
            done = true;
          }
          else {
            current_edge = current_edge.next;
          }
        }
      }
    }
  }
  
  static void _add_to_sbtree(_ScanBeamTreeEntries sbte, double y) {
    if ( sbte.sb_tree == null ) {
      // Add a new tree node here
      sbte.sb_tree = new _ScanBeamTree(y);
      sbte.sbt_entries++;
      return;
    }
     
    _ScanBeamTree tree_node = sbte.sb_tree;
    bool done = false;
    while(!done) {
      if (tree_node.y > y) {
        if (tree_node.less == null) {
          tree_node.less = new _ScanBeamTree(y);
          sbte.sbt_entries++;
          done = true;
        }
        else {
          tree_node = tree_node.less;
        }
      }
      else if (tree_node.y < y) {
        if (tree_node.more == null) {
          tree_node.more = new _ScanBeamTree(y);
          sbte.sbt_entries++;
          done = true;
        }
        else {
          tree_node = tree_node.more;
        }
      }
      else {
        done = true;
      }
    }
  }
  
  static LmtNode _bound_list(LmtTable lmt_table, double y) {
    if (lmt_table.top_node == null) {
      lmt_table.top_node = new LmtNode(y);
      return lmt_table.top_node;
    }
    else {
      LmtNode prev = null;
      LmtNode node = lmt_table.top_node;
      bool done = false;
      while (!done) {
        if (y < node.y) {
          // Insert a new LMT node before the current node
          LmtNode existing_node = node;
          node = new LmtNode(y);
          node.next = existing_node;
          if (prev == null) {
            lmt_table.top_node = node;
          }
          else {
            prev.next = node;
          }
          done = true;
        }
        else if (y > node.y) {
          // Head further up the LMT
          if (node.next == null) {
            node.next = new LmtNode(y);
            node = node.next;
            done = true;
          }
          else {
            prev = node;
            node = node.next;
          }
        }
        else {
          // Use this existing LMT node
          done = true;
        }
      }
      return node;
    }
  }

  static void _insert_bound(LmtNode lmt_node, _EdgeNode e) {
    if (lmt_node.first_bound == null) {
      // Link node e to the tail of the list
      lmt_node.first_bound = e ;
    }
    else {
      bool done = false;
      _EdgeNode prev_bound = null;
      _EdgeNode current_bound = lmt_node.first_bound;

      while (!done) {
        // Do primary sort on the x field
        if (e.bot.x < current_bound.bot.x) {
          // Insert a new node mid-list
          if (prev_bound == null) {
            lmt_node.first_bound = e;
          }
          else {
            prev_bound.next_bound = e;
          }
          e.next_bound = current_bound;
          
          done = true;
        }
        else if (e.bot.x == current_bound.bot.x) {
          // Do secondary sort on the dx field
          if (e.dx < current_bound.dx) {
            // Insert a new node mid-list
            if( prev_bound == null )
            {
              lmt_node.first_bound = e;
            }
            else
            {
              prev_bound.next_bound = e;
            }
            e.next_bound = current_bound;
            done = true;
          }
          else {
            // Head further down the list
            if( current_bound.next_bound == null ) {
              current_bound.next_bound = e;
              done = true;
            }
            else {
              prev_bound = current_bound ;
              current_bound = current_bound.next_bound;
            }
          }
        }
        else {
          // Head further down the list
          if( current_bound.next_bound == null ) {
            current_bound.next_bound = e;
            done = true;
          }
          else {
            prev_bound = current_bound ;
            current_bound = current_bound.next_bound;
          }
        }
      }
    }
  }
  
  static List<GRectangle> _create_contour_bboxes(Polygon p) {
    List<GRectangle> box = new List<GRectangle>(p.numberOfPolygons);
    
     /* Construct contour bounding boxes */
    for ( int c= 0; c < p.numberOfPolygons; c++) {
      Polygon inner_poly = p.getPolygon(c);
      box[c] = inner_poly.bounds;
    }
    return box;  
  }

  /**
   * // [type] = SUBJ/CLIP
   */
  static _EdgeTable _build_lmt(LmtTable lmt_table, _ScanBeamTreeEntries sbte, Polygon p, int type, OperationType operation)
  {
    // Create the entire input polygon edge table in one go
    _EdgeTable edge_table = new _EdgeTable();
    for (int c= 0; c < p.numberOfPolygons; c++) {
      Polygon ip = p.getPolygon(c);

      if (!ip.isContributing(0)) {
         // Ignore the non-contributing contour
         ip.setContributing(0, true);
      }
      else {
        // Perform contour optimisation
        int num_vertices= 0;
        int e_index = 0 ;
        edge_table = new _EdgeTable();
        for ( int i= 0; i < ip.numberOfPoints; i++) {
          if (OPTIMAL(ip, i) ) {
            GPoint pt = ip.getPoint(i);
            edge_table.addNode(pt.x, pt.y);
            // Record vertex in the scanbeam table
            _add_to_sbtree(sbte, pt.y);
            num_vertices++;
          }
        }
        
        // Do the contour forward pass
        for (int min= 0; min < num_vertices; min++) {
          // If a forward local minimum...
          if (edge_table.FWD_MIN(min)) {
            // Search for the next local maximum...
            int num_edges = 1;
            int max = NEXT_INDEX(min, num_vertices);
            while(edge_table.NOT_FMAX(max)) {
              num_edges++;
              max = NEXT_INDEX(max, num_vertices);
            }
            
            // Build the next edge list
            int v = min;
            _EdgeNode e = edge_table.getNode(e_index);
            e.bstate[BELOW] = BundleState.UNBUNDLED;
            e.bundle[BELOW][CLIP] = 0;
            e.bundle[BELOW][SUBJ] = 0;

            for (int i= 0; i < num_edges; i++) {
              _EdgeNode ei = edge_table.getNode(e_index+i);
              _EdgeNode ev = edge_table.getNode(v);
              
              ei.xb    = ev.vertex.x;
              ei.bot.x = ev.vertex.x;
              ei.bot.y = ev.vertex.y;
              
              v = NEXT_INDEX(v, num_vertices);
              ev = edge_table.getNode( v );
              
              ei.top.x= ev.vertex.x;
              ei.top.y= ev.vertex.y;
              ei.dx= (ev.vertex.x - ei.bot.x) / (ei.top.y - ei.bot.y);
              ei.type = type;
              ei.outp[ABOVE] = null ;
              ei.outp[BELOW] = null;
              ei.next = null;
              ei.prev = null;
              ei.succ = ((num_edges > 1) && (i < (num_edges - 1))) ? edge_table.getNode(e_index+i+1) : null;
              ei.next_bound = null ;
              ei.bside[CLIP] = (operation == OperationType.DIFF) ? RIGHT : LEFT;
              ei.bside[SUBJ] = LEFT ;
            }

            _insert_bound(_bound_list(lmt_table, edge_table.getNode(min).vertex.y), e);
            e_index += num_edges;
          }
        }
        
        // Do the contour reverse pass
        for (int min= 0; min < num_vertices; min++) {
          // If a reverse local minimum...
          if (edge_table.REV_MIN(min)) {
            // Search for the previous local maximum...
            int num_edges= 1;
            int max = PREV_INDEX(min, num_vertices);
            while(edge_table.NOT_RMAX(max)) {
               num_edges++;
               max = PREV_INDEX(max, num_vertices);
            }
            
            // Build the previous edge list
            int v = min;
            _EdgeNode e = edge_table.getNode(e_index);
            e.bstate[BELOW] = BundleState.UNBUNDLED;
            e.bundle[BELOW][CLIP] = 0;
            e.bundle[BELOW][SUBJ] = 0;

            for (int i= 0; i < num_edges; i++) {
              _EdgeNode ei = edge_table.getNode(e_index + i);
              _EdgeNode ev = edge_table.getNode(v);
              
              ei.xb    = ev.vertex.x;
              ei.bot.x = ev.vertex.x;
              ei.bot.y = ev.vertex.y;
              
              v = PREV_INDEX(v, num_vertices);
              ev = edge_table.getNode( v );
              
              ei.top.x = ev.vertex.x;
              ei.top.y = ev.vertex.y;
              ei.dx = (ev.vertex.x - ei.bot.x) / (ei.top.y - ei.bot.y);
              ei.type = type;
              ei.outp[ABOVE] = null;
              ei.outp[BELOW] = null;
              ei.next = null;
              ei.prev = null;
              ei.succ = ((num_edges > 1) && (i < (num_edges - 1))) ? edge_table.getNode(e_index+i+1) : null;
              //ei.pred = ((num_edges > 1) && (i > 0)) ? edge_table.getNode(e_index+i-1) : null ;
              ei.next_bound = null;
              ei.bside[CLIP] = (operation == OperationType.DIFF) ? RIGHT : LEFT;
              ei.bside[SUBJ] = LEFT;
            }
            
            _insert_bound(_bound_list(lmt_table, edge_table.getNode(min).vertex.y), e);
            e_index+= num_edges;
          }
        }
      }
    }
    
    return edge_table;
  }
  
  static void _minimax_test(Polygon subj, Polygon clip, OperationType operation) {
    List<GRectangle> s_bbox = _create_contour_bboxes(subj);
    List<GRectangle> c_bbox = _create_contour_bboxes(clip);
    
    int subj_num_poly = subj.numberOfPolygons;
    int clip_num_poly = clip.numberOfPolygons;
    
    List<List<bool>> o_table = new List<List<bool>>();
    // (subj_num_poly * clip_num_poly);
    for(int sb = 0; sb < subj_num_poly; sb++) {
      List<bool> clp = new List<bool>(clip_num_poly);
      o_table.add(clp);
    }
    
    // Check all subject contour bounding boxes against clip boxes
    for( int s = 0; s < subj_num_poly; s++ ) {
      for( int c = 0; c < clip_num_poly ; c++ ) {
        o_table[s][c] =
         (!((s_bbox[s].maxX < c_bbox[c].minX) ||
            (s_bbox[s].minX > c_bbox[c].maxX))) &&
         (!((s_bbox[s].maxY < c_bbox[c].minY) ||
            (s_bbox[s].minY > c_bbox[c].maxY)));
      }
    }
     // For each clip contour, search for any subject contour overlaps
    for(int c = 0; c < clip_num_poly; c++) {
      bool overlap = false;
      for( int s = 0; !overlap && (s < subj_num_poly) ; s++) {
         overlap = o_table[s][c];
      }
      if (!overlap) {
         clip.setContributing( c, false ); // Flag non contributing status
      }
    }  
    if (operation == OperationType.INT)
    {  
       /* For each subject contour, search for any clip contour overlaps */
       for ( int s= 0; s < subj_num_poly; s++)
       {
          bool overlap = false;
          for ( int c= 0; !overlap && (c < clip_num_poly); c++)
          {
             overlap = o_table[s][c];
          }
          if (!overlap)
          {
             subj.setContributing( s, false ); // Flag non contributing status
         }
       }  
    }
  }

  static Polygon _createNewPolygon(String classType) {
    if (classType == "SimplePolygon")
      return new SimplePolygon();
    else if (classType == "ComplexPolygon")
      return new ComplexPolygon();
    else throw new GeoClipException("Boolean: unknown polygon type.");
  }

  static Polygon clip(OperationType operation, Polygon subject, Polygon clip, String classType) {
    Polygon result = _createNewPolygon(classType);
    try {
      if ((subject.isEmpty && clip.isEmpty) || (subject.isEmpty && ((operation ==
          OperationType.INT) || (operation == OperationType.DIFF))) || (clip.isEmpty &&
          (operation == OperationType.INT))) {
        return result;
      }
  
      // Identify potentially contributing contours
      if (((operation == OperationType.INT) || (operation == OperationType.DIFF))
          && !subject.isEmpty && !clip.isEmpty) {
        _minimax_test(subject, clip, operation);
      }
  
      // Build LMT
      LmtTable lmt_table = new LmtTable();
      _ScanBeamTreeEntries sbte = new _ScanBeamTreeEntries();
      if (!subject.isEmpty) {
        _build_lmt(lmt_table, sbte, subject, SUBJ, operation);
      }
  
      if (!clip.isEmpty) {
        _build_lmt(lmt_table, sbte, clip, CLIP, operation);
      }
  
      // Return a NULL result if no contours contribute
      if (lmt_table.top_node == null) {
        return result;
      }
  
      // Build scanbeam table from scanbeam tree
      List<double> sbt = sbte.build_sbt();
  
      List<int> parity = new List<int>(2);
      parity[0] = LEFT;
      parity[1] = LEFT;
  
      // Invert clip polygon for difference operation
      if (operation == OperationType.DIFF) {
        parity[CLIP] = RIGHT;
      }
  
      LmtNode local_min = lmt_table.top_node;
  
      // used to create resulting Polygon
      _TopPolygonNode out_poly = new _TopPolygonNode();
  
      _AetTree aet = new _AetTree();
      int scanbeam = 0;
  
      // Scan process BEGIN
      // Process each scanbeam
      while (scanbeam < sbt.length) {
        // Set yb and yt to the bottom and top of the scanbeam
        double yb = sbt[scanbeam++];
        double yt = 0.0;
        double dy = 0.0;
        if (scanbeam < sbt.length) {
          yt = sbt[scanbeam];
          dy = yt - yb;
        }
        
        // Scanbeam boundary processing...
        // If LMT node corresponding to yb exists
        if (local_min != null) {
          if (local_min.y == yb) {
            /* Add edges starting at this local minimum to the AET */
            for(_EdgeNode edge = local_min.first_bound; (edge != null) ; edge= edge.next_bound) {
              _add_edge_to_aet(aet, edge);
            }
  
            local_min = local_min.next;
          }
        }
  
        /* Set dummy previous x value */
        double px = -double.MAX_FINITE;
        
        /* Create bundles within AET */
        _EdgeNode e0 = aet.top_node;
        _EdgeNode e1 = aet.top_node;
  
        /* Set up bundle fields of first edge */
        aet.top_node.bundle[ABOVE][ aet.top_node.type ] = (aet.top_node.top.y != yb) ? 1 : 0;
        aet.top_node.bundle[ABOVE][ ((aet.top_node.type==0) ? 1 : 0) ] = 0;
        aet.top_node.bstate[ABOVE] = BundleState.UNBUNDLED;
  
        for (_EdgeNode next_edge= aet.top_node.next ; (next_edge != null); next_edge = next_edge.next) {
          int ne_type = next_edge.type ;
          int ne_type_opp = ((next_edge.type==0) ? 1 : 0); //next edge type opposite
          
          /* Set up bundle fields of next edge */
          next_edge.bundle[ABOVE][ ne_type     ]= (next_edge.top.y != yb) ? 1 : 0;
          next_edge.bundle[ABOVE][ ne_type_opp ] = 0 ;
          next_edge.bstate[ABOVE] = BundleState.UNBUNDLED;
          
          /* Bundle edges above the scanbeam boundary if they coincide */
          if (next_edge.bundle[ABOVE][ne_type] == 1) {
            if (EQ(e0.xb, next_edge.xb) && EQ(e0.dx, next_edge.dx) && (e0.top.y != yb)) {
              next_edge.bundle[ABOVE][ ne_type     ] ^= e0.bundle[ABOVE][ ne_type     ];
              next_edge.bundle[ABOVE][ ne_type_opp ]  = e0.bundle[ABOVE][ ne_type_opp ];
              next_edge.bstate[ABOVE] = BundleState.BUNDLE_HEAD;
              e0.bundle[ABOVE][CLIP] = 0;
              e0.bundle[ABOVE][SUBJ] = 0;
              e0.bstate[ABOVE] = BundleState.BUNDLE_TAIL;
            }
            e0 = next_edge;
          }
        }
  
        List<int> horiz = new List<int>(2);
        horiz[CLIP]= _HState.NH;
        horiz[SUBJ]= _HState.NH;
        
        List<int> exists = new List<int>(2);
        exists[CLIP] = 0;
        exists[SUBJ] = 0;
        
        _PolygonNode cf;
        
        // Process each edge at this scanbeam boundary
        for (_EdgeNode edge= aet.top_node ; (edge != null); edge = edge.next ) {
          exists[CLIP] = edge.bundle[ABOVE][CLIP] + (edge.bundle[BELOW][CLIP] << 1);
          exists[SUBJ] = edge.bundle[ABOVE][SUBJ] + (edge.bundle[BELOW][SUBJ] << 1);
  
          if ((exists[CLIP] != 0) || (exists[SUBJ] != 0)) {
            // Set bundle side
            edge.bside[CLIP] = parity[CLIP];
            edge.bside[SUBJ] = parity[SUBJ];
            
            bool contributing = false;
            int br=0, bl=0, tr=0, tl=0;
  
            /* Determine contributing status and quadrant occupancies */
            if ((operation == OperationType.DIFF) || (operation == OperationType.INT)) {
              contributing= ((exists[CLIP]!=0) && ((parity[SUBJ]!=0) || (horiz[SUBJ]!=0))) ||
                            ((exists[SUBJ]!=0) && ((parity[CLIP]!=0) || (horiz[CLIP]!=0))) ||
                            ((exists[CLIP]!=0) && (exists[SUBJ]!=0) && (parity[CLIP] == parity[SUBJ]));
              br = ((parity[CLIP]!=0) && (parity[SUBJ]!=0)) ? 1 : 0;
              bl = ( ((parity[CLIP] ^ edge.bundle[ABOVE][CLIP])!=0) &&
                    ((parity[SUBJ] ^ edge.bundle[ABOVE][SUBJ])!=0) ) ? 1 : 0;
              tr = ( ((parity[CLIP] ^ ((horiz[CLIP]!=_HState.NH)?1:0)) !=0) && 
                     ((parity[SUBJ] ^ ((horiz[SUBJ]!=_HState.NH)?1:0)) !=0) ) ? 1 : 0;
              tl = (((parity[CLIP] ^ ((horiz[CLIP]!=_HState.NH)?1:0) ^ edge.bundle[BELOW][CLIP])!=0) &&
                    ((parity[SUBJ] ^ ((horiz[SUBJ]!=_HState.NH)?1:0) ^ edge.bundle[BELOW][SUBJ])!=0))?1:0;
            }
            else if (operation == OperationType.XOR) {
              contributing= (exists[CLIP]!=0) || (exists[SUBJ]!=0);
              br= (parity[CLIP]) ^ (parity[SUBJ]);
              bl= (parity[CLIP] ^ edge.bundle[ABOVE][CLIP]) ^ (parity[SUBJ] ^ edge.bundle[ABOVE][SUBJ]);
              tr= (parity[CLIP] ^ ((horiz[CLIP]!=_HState.NH)?1:0)) ^ (parity[SUBJ] ^ ((horiz[SUBJ]!=_HState.NH)?1:0));
              tl= (parity[CLIP] ^ ((horiz[CLIP]!=_HState.NH)?1:0) ^ edge.bundle[BELOW][CLIP])
                ^ (parity[SUBJ] ^ ((horiz[SUBJ]!=_HState.NH)?1:0) ^ edge.bundle[BELOW][SUBJ]);
            }
            else if (operation == OperationType.UNION ) {
              contributing= ((exists[CLIP]!=0) && (!(parity[SUBJ]!=0) || (horiz[SUBJ]!=0))) ||
                            ((exists[SUBJ]!=0) && (!(parity[CLIP]!=0) || (horiz[CLIP]!=0))) ||
                            ((exists[CLIP]!=0) && (exists[SUBJ]!=0) && (parity[CLIP] == parity[SUBJ]));
              br= ((parity[CLIP]!=0) || (parity[SUBJ]!=0))?1:0;
              bl= (((parity[CLIP] ^ edge.bundle[ABOVE][CLIP])!=0) || ((parity[SUBJ] ^ edge.bundle[ABOVE][SUBJ])!=0))?1:0;
              tr= ( ((parity[CLIP] ^ ((horiz[CLIP]!=_HState.NH)?1:0))!=0) || 
                    ((parity[SUBJ] ^ ((horiz[SUBJ]!=_HState.NH)?1:0))!=0) ) ?1:0;
              tl= ( ((parity[CLIP] ^ ((horiz[CLIP]!=_HState.NH)?1:0) ^ edge.bundle[BELOW][CLIP])!=0) ||
                    ((parity[SUBJ] ^ ((horiz[SUBJ]!=_HState.NH)?1:0) ^ edge.bundle[BELOW][SUBJ])!=0) ) ? 1:0;
            }
            else {
               throw new GeoClipException("Unknown op");
            }
            
            // Update parity
            parity[CLIP] ^= edge.bundle[ABOVE][CLIP];
            parity[SUBJ] ^= edge.bundle[ABOVE][SUBJ];
            
            // Update horizontal state
            if (exists[CLIP]!=0) {
               horiz[CLIP] = _HState.next_h_state[horiz[CLIP]][((exists[CLIP] - 1) << 1) + parity[CLIP]];
            }
            if ( exists[SUBJ]!=0) {
               horiz[SUBJ] = _HState.next_h_state[horiz[SUBJ]][((exists[SUBJ] - 1) << 1) + parity[SUBJ]];
            }
            
            if (contributing) {
              double xb = edge.xb;
               
              int vclass = _VertexType.getType(tr, tl, br, bl);
              
              switch (vclass) {
                case _VertexType.EMN:
                case _VertexType.IMN:
                  edge.outp[ABOVE] = out_poly.add_local_min(xb, yb);
                  px = xb;
                  cf = edge.outp[ABOVE];
                  break;
                case _VertexType.ERI:
                   if (xb != px) {
                     cf.add_right( xb, yb);
                     px = xb;
                   }
                   edge.outp[ABOVE]= cf;
                   cf = null;
                   break;
                case _VertexType.ELI:
                  edge.outp[BELOW].add_left( xb, yb);
                  px= xb;
                  cf = edge.outp[BELOW];
                  break;
                case _VertexType.EMX:
                  if (xb != px) {
                    cf.add_left( xb, yb);
                    px= xb;
                  }
                  out_poly.merge_right(cf, edge.outp[BELOW]);
                  cf = null;
                  break;
                case _VertexType.ILI:
                  if (xb != px) {
                    cf.add_left( xb, yb);
                    px= xb;
                  }
                  edge.outp[ABOVE]= cf;
                  cf = null;
                  break;
                case _VertexType.IRI:
                  edge.outp[BELOW].add_right( xb, yb );
                  px= xb;
                  cf = edge.outp[BELOW];
                  edge.outp[BELOW]= null;
                  break;
                case _VertexType.IMX:
                  if (xb != px) {
                    cf.add_right( xb, yb );
                    px= xb;
                  }
                  out_poly.merge_left(cf, edge.outp[BELOW]);
                  cf = null;
                  edge.outp[BELOW]= null;
                  break;
                case _VertexType.IMM:
                  if (xb != px) {
                    cf.add_right( xb, yb);
                    px= xb;
                  }
                  out_poly.merge_left(cf, edge.outp[BELOW]);
                  edge.outp[BELOW]= null;
                  edge.outp[ABOVE] = out_poly.add_local_min(xb, yb);
                  cf = edge.outp[ABOVE];
                  break;
                case _VertexType.EMM:
                  if (xb != px) {
                    cf.add_left( xb, yb);
                    px= xb;
                  }
                  out_poly.merge_right(cf, edge.outp[BELOW]);
                  edge.outp[BELOW]= null;
                  edge.outp[ABOVE] = out_poly.add_local_min(xb, yb);
                  cf = edge.outp[ABOVE];
                  break;
                case _VertexType.LED:
                  // TODO edge.outp has nothing for 866.0,640.0 
                  if (edge.bot.y == yb)
                    edge.outp[BELOW].add_left( xb, yb);
                  edge.outp[ABOVE]= edge.outp[BELOW];
                  px= xb;
                  break;
                case _VertexType.RED:
                  if (edge.bot.y == yb)
                    edge.outp[BELOW].add_right( xb, yb );
                  edge.outp[ABOVE]= edge.outp[BELOW];
                  px= xb;
                  break;
                default:
                  break;
              } // end switch
            } // end if contributing
          } // End of edge exists conditional
        } // End of AET loop
        
        // Delete terminating edges from the AET, otherwise compute xt
        for (_EdgeNode edge = aet.top_node ; (edge != null); edge = edge.next) {
          if (edge.top.y == yb) {
            _EdgeNode prev_edge = edge.prev;
            _EdgeNode next_edge= edge.next;
            
            if (prev_edge != null)
              prev_edge.next = next_edge;
            else
              aet.top_node = next_edge;
            
            if (next_edge != null )
              next_edge.prev = prev_edge;
            
            // Copy bundle head state to the adjacent tail edge if required
            if ((edge.bstate[BELOW] == BundleState.BUNDLE_HEAD) && (prev_edge!=null)) {
              if (prev_edge.bstate[BELOW] == BundleState.BUNDLE_TAIL) {
                prev_edge.outp[BELOW]= edge.outp[BELOW];
                prev_edge.bstate[BELOW]= BundleState.UNBUNDLED;
                if ( prev_edge.prev != null) {
                  if (prev_edge.prev.bstate[BELOW] == BundleState.BUNDLE_TAIL) {
                    prev_edge.bstate[BELOW] = BundleState.BUNDLE_HEAD;
                  }
                }
              }
            }
          }
          else {
            if (edge.top.y == yt)
              edge.xt= edge.top.x;
            else
              edge.xt= edge.bot.x + edge.dx * (yt - edge.bot.y);
          }
        } // End of delete terminating edges loop
        
        // SCAN BEAM INTERIOR PROCESSING
        if (scanbeam < sbte.sbt_entries) {
          // Build intersection table for the current scanbeam
          _ItNodeTable it_table = new _ItNodeTable();
          it_table.build_intersection_table(aet, dy);
          
          // Process each node in the intersection table
          for (_ItNode intersect = it_table.top_node ; (intersect != null); intersect = intersect.next) {
            e0= intersect.ie[0];
            e1= intersect.ie[1];
  
            // Only generate output for contributing intersections
            if ( ((e0.bundle[ABOVE][CLIP]!=0) || (e0.bundle[ABOVE][SUBJ]!=0)) &&
                 ((e1.bundle[ABOVE][CLIP]!=0) || (e1.bundle[ABOVE][SUBJ]!=0))) {
              _PolygonNode p = e0.outp[ABOVE];
              _PolygonNode q = e1.outp[ABOVE];
              double ix = intersect.point.x;
              double iy = intersect.point.y + yb;
  
              int in_clip = ( ( (e0.bundle[ABOVE][CLIP]!=0) && !(e0.bside[CLIP]!=0)) ||
                              ( (e1.bundle[ABOVE][CLIP]!=0) &&  (e1.bside[CLIP]!=0)) ||
                              (!(e0.bundle[ABOVE][CLIP]!=0) && !(e1.bundle[ABOVE][CLIP]!=0) &&
                                (e0.bside[CLIP]!=0) && (e1.bside[CLIP]!=0) ) ) ? 1 : 0;
              
              int in_subj = ( ( (e0.bundle[ABOVE][SUBJ]!=0) && !(e0.bside[SUBJ]!=0)) ||
                              ( (e1.bundle[ABOVE][SUBJ]!=0) &&  (e1.bside[SUBJ]!=0)) ||
                              (!(e0.bundle[ABOVE][SUBJ]!=0) && !(e1.bundle[ABOVE][SUBJ]!=0) &&
                                (e0.bside[SUBJ]!=0) && (e1.bside[SUBJ]!=0) ) ) ? 1 : 0;
  
              int tr=0, tl=0, br=0, bl=0 ;
              // Determine quadrant occupancies
              if( (operation == OperationType.DIFF) || (operation == OperationType.INT) ) {
                tr= ((in_clip!=0) && (in_subj!=0)) ? 1 : 0;
                tl= (((in_clip ^ e1.bundle[ABOVE][CLIP])!=0) && ((in_subj ^ e1.bundle[ABOVE][SUBJ])!=0))?1:0;
                br= (((in_clip ^ e0.bundle[ABOVE][CLIP])!=0) && ((in_subj ^ e0.bundle[ABOVE][SUBJ])!=0))?1:0;
                bl= (((in_clip ^ e1.bundle[ABOVE][CLIP] ^ e0.bundle[ABOVE][CLIP])!=0) &&
                     ((in_subj ^ e1.bundle[ABOVE][SUBJ] ^ e0.bundle[ABOVE][SUBJ])!=0) ) ? 1:0;
              }
              else if( operation == OperationType.XOR ) {
                tr= (in_clip)^ (in_subj);
                tl= (in_clip ^ e1.bundle[ABOVE][CLIP]) ^ (in_subj ^ e1.bundle[ABOVE][SUBJ]);
                br= (in_clip ^ e0.bundle[ABOVE][CLIP]) ^ (in_subj ^ e0.bundle[ABOVE][SUBJ]);
                bl= (in_clip ^ e1.bundle[ABOVE][CLIP] ^ e0.bundle[ABOVE][CLIP])
                  ^ (in_subj ^ e1.bundle[ABOVE][SUBJ] ^ e0.bundle[ABOVE][SUBJ]);
              }
              else if( operation == OperationType.UNION ) {
                tr= ((in_clip!=0) || (in_subj!=0)) ? 1 : 0;
                tl= (((in_clip ^ e1.bundle[ABOVE][CLIP])!=0) || ((in_subj ^ e1.bundle[ABOVE][SUBJ])!=0)) ? 1 : 0;
                br= (((in_clip ^ e0.bundle[ABOVE][CLIP])!=0) || ((in_subj ^ e0.bundle[ABOVE][SUBJ])!=0)) ? 1 : 0;
                bl= (((in_clip ^ e1.bundle[ABOVE][CLIP] ^ e0.bundle[ABOVE][CLIP])!=0) ||
                     ((in_subj ^ e1.bundle[ABOVE][SUBJ] ^ e0.bundle[ABOVE][SUBJ])!=0)) ? 1 : 0;
              }
              else {
                throw new GeoClipException("Unknown op type, $operation");
              }
              
              int vclass = _VertexType.getType(tr, tl, br, bl);
  
              switch (vclass) {
                case _VertexType.EMN:
                  e0.outp[ABOVE] = out_poly.add_local_min(ix, iy);
                  e1.outp[ABOVE] = e0.outp[ABOVE];
                  break;
                case _VertexType.ERI:
                  if (p != null) {
                    p.add_right(ix, iy);
                    e1.outp[ABOVE]= p;
                    e0.outp[ABOVE]= null;
                  }
                  break;
                case _VertexType.ELI:
                  if (q != null) {
                    q.add_left(ix, iy);
                    e0.outp[ABOVE]= q;
                    e1.outp[ABOVE]= null;
                  }
                  break;
                case _VertexType.EMX:
                  if ((p!=null) && (q!=null)) {
                    p.add_left( ix, iy);
                    out_poly.merge_right(p, q);
                    e0.outp[ABOVE]= null;
                    e1.outp[ABOVE]= null;
                  }
                  break;
                case _VertexType.IMN:
                  e0.outp[ABOVE] = out_poly.add_local_min(ix, iy);
                  e1.outp[ABOVE]= e0.outp[ABOVE];
                  break;
                case _VertexType.ILI:
                  if (p != null) {
                    p.add_left(ix, iy);
                    e1.outp[ABOVE]= p;
                    e0.outp[ABOVE]= null;
                  }
                  break;
                case _VertexType.IRI:
                  if (q!=null) {
                    q.add_right(ix, iy);
                    e0.outp[ABOVE]= q;
                    e1.outp[ABOVE]= null;
                  }
                  break;
                case _VertexType.IMX:
                  if ((p!=null) && (q!=null)) {
                    p.add_right(ix, iy);
                    out_poly.merge_left(p, q);
                    e0.outp[ABOVE]= null;
                    e1.outp[ABOVE]= null;
                  }
                  break;
                case _VertexType.IMM:
                  if ((p!=null) && (q!=null)) {
                    p.add_right(ix, iy);
                    out_poly.merge_left(p, q);
                    e0.outp[ABOVE] = out_poly.add_local_min(ix, iy);
                    e1.outp[ABOVE]= e0.outp[ABOVE];
                  }
                  break;
                case _VertexType.EMM:
                  if ((p!=null) && (q!=null)) {
                    p.add_left(ix, iy);
                    out_poly.merge_right(p, q);
                    e0.outp[ABOVE] = out_poly.add_local_min(ix, iy);
                    e1.outp[ABOVE] = e0.outp[ABOVE];
                  }
                  break;
                default:
                   break;
              } // End of switch
            } // End of contributing intersection conditional
  
            // Swap bundle sides in response to edge crossing
            if (e0.bundle[ABOVE][CLIP]!=0)
              e1.bside[CLIP] = (e1.bside[CLIP]==0)?1:0;
            if (e1.bundle[ABOVE][CLIP]!=0)
              e0.bside[CLIP]= (e0.bside[CLIP]==0)?1:0;
            if (e0.bundle[ABOVE][SUBJ]!=0)
              e1.bside[SUBJ]= (e1.bside[SUBJ]==0)?1:0;
            if (e1.bundle[ABOVE][SUBJ]!=0)
              e0.bside[SUBJ]= (e0.bside[SUBJ]==0)?1:0;
            
            // Swap e0 and e1 bundles in the AET
            _EdgeNode prev_edge = e0.prev;
            _EdgeNode next_edge = e1.next;
            if (next_edge != null) {
              next_edge.prev = e0;
            }
  
            if (e0.bstate[ABOVE] == BundleState.BUNDLE_HEAD) {
              bool search = true;
              while (search) {
                prev_edge= prev_edge.prev;
                if (prev_edge != null) {
                  if (prev_edge.bstate[ABOVE] != BundleState.BUNDLE_TAIL) {
                    search= false;
                  }
                }
                else {
                  search= false;
                }
              }
            }
  
            if (prev_edge == null) {
              aet.top_node.prev = e1;
              e1.next = aet.top_node;
              aet.top_node = e0.next;
            }
            else {
              prev_edge.next.prev = e1;
              e1.next = prev_edge.next;
              prev_edge.next = e0.next;
            }
  
            e0.next.prev = prev_edge;
            e1.next.prev = e1;
            e0.next = next_edge;
          } // End of IT loop
          
          // Prepare for next scanbeam
          for (_EdgeNode edge = aet.top_node; (edge != null); edge = edge.next) {
            _EdgeNode next_edge = edge.next;
            _EdgeNode succ_edge = edge.succ;
            if ((edge.top.y == yt) && (succ_edge!=null)) {
              // Replace AET edge by its successor
              succ_edge.outp[BELOW]= edge.outp[ABOVE];
              succ_edge.bstate[BELOW]= edge.bstate[ABOVE];
              succ_edge.bundle[BELOW][CLIP]= edge.bundle[ABOVE][CLIP];
              succ_edge.bundle[BELOW][SUBJ]= edge.bundle[ABOVE][SUBJ];
              _EdgeNode prev_edge = edge.prev;
              if ( prev_edge != null )
                prev_edge.next = succ_edge;
              else
                aet.top_node = succ_edge;
              if (next_edge != null)
                next_edge.prev= succ_edge;
              succ_edge.prev = prev_edge;
              succ_edge.next = next_edge;
            }
            else {
              // Update this edge
              edge.outp[BELOW]= edge.outp[ABOVE];
              edge.bstate[BELOW]= edge.bstate[ABOVE];
              edge.bundle[BELOW][CLIP]= edge.bundle[ABOVE][CLIP];
              edge.bundle[BELOW][SUBJ]= edge.bundle[ABOVE][SUBJ];
              edge.xb= edge.xt;
            }
            edge.outp[ABOVE]= null;
          } // End of next scan beam loop
        } // End of Scan beam interior processing
      } // End while -- Scan process
      
      // Generate result polygon from out_poly
      result = out_poly.getResult(classType);
    }
    on NoSuchMethodError catch(e) {
      throw new GeoClipException("Boolean.clip: [$e] -- ${e.stackTrace}");
    }
    catch(e) {
      throw new GeoClipException("Boolean.clip: unknown exception [$e] -- ${e.stackTrace}");
    }
    
    return result;
  }
}
