part of geoclip;

/**
 * Horizontal edge states            
 */
class _HState {
   static final int NH = 0 ; /* No horizontal edge                */
   static final int BH = 1 ; /* Bottom horizontal edge            */
   static final int TH = 2 ; /* Top horizontal edge               */

   /* Horizontal edge state transitions within scanbeam boundary */
   static final List<List<int>> next_h_state =
   [
   /*        ABOVE     BELOW     CROSS */
   /*        L   R     L   R     L   R */  
   /* NH */ [BH, TH,   TH, BH,   NH, NH],
   /* BH */ [NH, NH,   NH, NH,   TH, TH],
   /* TH */ [NH, NH,   NH, NH,   BH, BH]
   ];
}

/**
 * Edge bundle state                 
 */
class BundleState {
  String _State;
  BundleState(this._State);

  // Isolated edge not within a bundle
  static final BundleState UNBUNDLED = new BundleState("UNBUNDLED");
  // Bundle head node
  static final BundleState BUNDLE_HEAD = new BundleState("BUNDLE_HEAD");
  // Passive bundle tail node
  static final BundleState BUNDLE_TAIL = new BundleState("BUNDLE_TAIL");

  @override
  String toString() {
    return _State;
  }
}

class _EdgeNode {
  // Piggy-backed contour vertex data
  GPoint vertex = new GPoint.zero();
  // Edge lower (x, y) coordinate
  GPoint bot = new GPoint.zero();
  // Edge upper (x, y) coordinate
  GPoint top = new GPoint.zero();
  // Scanbeam bottom x coordinate
  double xb;
  // Scanbeam top x coordinate
  double xt;
  // Change in x for a unit y increase
  double dx;
  // Clip / subject edge flag
  int type;
  // [2][2] Bundle edge flags
  List<List<int>> bundle = new List<List<int>>(2);
  // [2] Bundle left / right indicators
  List<int> bside = new List<int>(2);
  // [2] Edge bundle state
  List<BundleState> bstate = new List<BundleState>(2);
  // [2] Output polygon / tristrip pointer
  List<_PolygonNode> outp = new List<_PolygonNode>(2);
  // Previous edge in the AET
  _EdgeNode prev;
  // Next edge in the AET
  _EdgeNode next;
  //EdgeNode       pred;         // Edge connected at the lower end
  // Edge connected at the upper end
  _EdgeNode succ;
  // Pointer to next bound in LMT
  _EdgeNode next_bound;
  
  _EdgeNode() {
    // [2][2]
    for(int r = 0; r < 2; r++) {
      List<int> i = new List<int>(2);
      bundle[r] = i;
    }

  }
}

class _EdgeTable
{
   List<_EdgeNode> _List = new List<_EdgeNode>();

   void addNode( double x, double y ) {
     _EdgeNode node = new _EdgeNode();
     node.vertex.x = x ;
     node.vertex.y = y ;
     _List.add( node );
   }
   
   _EdgeNode getNode( int index ) {
     return _List[index];
   }
   
   bool FWD_MIN(int i) {
     _EdgeNode prev = _List[Boolean.PREV_INDEX(i, _List.length)];
     _EdgeNode next = _List[Boolean.NEXT_INDEX(i, _List.length)];
     _EdgeNode ith  = _List[i];
     return ((prev.vertex.y >= ith.vertex.y) &&
             (next.vertex.y >  ith.vertex.y));
   }

   bool NOT_FMAX(int i) {
     _EdgeNode next = _List[Boolean.NEXT_INDEX(i, _List.length)];
     _EdgeNode ith  = _List[i];
     return(next.vertex.y > ith.vertex.y);
   }

   bool REV_MIN(int i) {
     _EdgeNode prev = _List[Boolean.PREV_INDEX(i, _List.length)];
     _EdgeNode next = _List[Boolean.NEXT_INDEX(i, _List.length)];
     _EdgeNode ith  = _List[i];
     return ((prev.vertex.y >  ith.vertex.y) &&
             (next.vertex.y >= ith.vertex.y));
   }
   
   bool NOT_RMAX(int i) {
     _EdgeNode prev = _List[Boolean.PREV_INDEX(i, _List.length)];
     _EdgeNode ith  = _List[i];
     return (prev.vertex.y > ith.vertex.y) ;
   }
}
