part of geoclip;

abstract class Polygon {
  bool _isHole = false;
  bool _contributes = false;
  
   /**
    * Flush all points/vertices.
    */
   void clear();
   
   /**
    * Add a vertex/coord to [Polygon].
    */
   void addCoord(double x, double y);
   
   /**
    * Add a vertex/coord to [Polygon].
    */
   void addPoint(GPoint p);
   
   /**
    * Add [p] to this [Polygon] 
    * [p] should not contain other [Polygon]s.
    */
   void addPolygon(Polygon p) {
     
   }
   
   /**
    * Returns [true] if the [Polygon] not zero points/vertices.
    */
   bool get isEmpty;
   
   /**
    * Returns the bounding rectangle of this polygon.
    */
   GRectangle get bounds;
   
   /**
    * Returns a [Polygon] embedded within this [Polygon].
    */
   Polygon getPolygon(int polyIndex);
   
   /**
    * Returns the number of [Polygon]s embedded.
    */
   int get numberOfPolygons;
   
   /**
    * Return the number points of outer [Polygon].
    */
   int get numberOfPoints;
   
   /**
    * Return the point component of outer polygon.
    */
   GPoint getPoint(int index);
   
   /**
    * Return true if this [Polygon] is an embedded polygon.
    * The embedded polygon is consider to be/form a "hole".
    */
   bool get isHole => _isHole;
   
   /**
    * Set whether or not this polygon is a hole.
    * This is invalid on [SimplePolygon]s.
    * 
    * TODO @throws GeoClipException if called on a complex polygon.
    */
   set isHole(bool isHole) => _isHole = isHole;
   
   /**
    * Return [true] if the embedded polygon is contributing to the boolean
    * operation.
    */
   bool isContributing(int index);
   
   /**
    * Set whether or not this inner polygon is constributing to the set operation.
    * This method should NOT be used outside the Clip algorithm.
    */
   void setContributing(int index, bool contributes);
   
   /**
    * Return a [Polygon] that is the Intersection of this [Polygon] with the given polygon.
    */
   Polygon intersection(Polygon p);
      
   /**
    * Return a [Polygon] that is the Union of this [Polygon] with the given polygon.
    */
   Polygon add(Polygon p);
   
   /**
    * Return a [Polygon] that is the exclusive-or of this [Polygon] with the given polygon.
    */
   Polygon xor(Polygon p);
   
   /**
    * Return a [Polygon] that is the difference of this [Polygon] with the given polygon.
    */
   Polygon subtract(Polygon p);
   
   /**
    * Return the Area of the [Polygon] in Square-Units.
    */
   double get area;
}
