part of geoclip;

class SimplePolygon extends Polygon {
  List<GPoint> _List = new List<GPoint>();
  
  bool _Contributes = true;
   
  SimplePolygon();
   
   /**
    * [true] if the given object is equal to this one.
    * WARNING: [false] if the first point occurs more than once in the list.
    */
  bool operator==(Polygon other) {
    if (!(other is SimplePolygon))
      return false;
      
    SimplePolygon that = other as SimplePolygon;
    
    int this_num = _List.length;
    int that_num = that._List.length;
    if ( this_num != that_num )
      return false ;
    
    if (this_num > 0) {
      GPoint this_p = getPoint(0);
      int that_first_index = -1;
      for(int that_index = 0; (that_first_index == -1) && (that_index < that_num) ; that_index++) {
        GPoint that_p = that.getPoint(that_index);
        if ((this_p.x == that_p.x) && (this_p.y == that_p.y)) {
          that_first_index = that_index;
        }
      }
       
      if (that_first_index == -1)
        return false;
      
      int that_index = that_first_index;
      for( int this_index = 0 ; this_index < this_num ; this_index++ ) {
        GPoint this_p = getPoint(this_index);
        GPoint that_p = that.getPoint(that_index);
        
        if ( (this_p.x != that_p.x) || (this_p.y != that_p.y) )
          return false;
           
        that_index++ ;
        if ( that_index >= that_num ) {
           that_index = 0;
        }
      }
    }
    return true;
  }
   
  //   /**
  //    * Return the hashCode of the object.
  //    * <p>
  //    * <strong>WARNING:</strong>Hash and Equals break contract.
  //    *
  //    * @return an integer value that is the same for two objects
  //    * whenever their internal representation is the same (equals() is true)
  //    */
  //   int hashCode()
  //   {
  //      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //      // !!! WARNING:  This hash and equals break the contract. !!!
  //      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //      int result = 17 ;
  //      result = 37*result + m_List.hashCode();
  //      return result;
  //   }
   
  String toString() {
    return "SimplePolygon: points=${numberOfPoints}";
  }
   
   /**
    * Reset polygon.
    */
  void clear() {
    _List.clear();
  }
   
  /**
    * Add [GPoint] to outer [Polygon]
    */
  void addCoord(double x, double y) {
    addPoint(new GPoint(x, y));
  }
   
   /**
    * Add [GPoint] to [Polygon]
    */
   void addPoint(GPoint p) {
      _List.add(p);
   }
  
   /**
    * Throws GeoClipException.
    */
   void addPolygon(Polygon p) {
      throw new GeoClipException("Cannot add polygon to a SimplePolygon.");
   }
   
   /**
    * Return true if the polygon is empty
    */
   bool get isEmpty => _List.isEmpty;
   
   /**
    * Returns the bounding rectangle of this polygon.
    */
   GRectangle get bounds {
     double xmin = double.MAX_FINITE;
     double ymin = double.MAX_FINITE;
     double xmax = -double.MAX_FINITE;
     double ymax = -double.MAX_FINITE;
      
     for(GPoint p in _List) {
       if( p.x < xmin ) xmin = p.x;
       if( p.x > xmax ) xmax = p.x;
       if( p.y < ymin ) ymin = p.y;
       if( p.y > ymax ) ymax = p.y;
     }
  //      for( int i = 0 ; i < m_List.size() ; i++ ) {
  //         double x = getX(i);
  //         double y = getY(i);
  //         if( x < xmin ) xmin = x;
  //         if( x > xmax ) xmax = x;
  //         if( y < ymin ) ymin = y;
  //         if( y > ymax ) ymax = y;
  //      }
      
     return new GRectangle(xmin, ymin, (xmax-xmin), (ymax-ymin));
   }
   
   /**
    * throws GeoClipException.
    */
   Polygon getPolygon(int polyIndex) {
      if (polyIndex != 0) {
         throw new GeoClipException("SimplePolygon only has one polygon");
      }
      return this;
   }
   
   /**
    * Always 1.
    */
   int get numberOfPolygons => 1;
   
   /**
    * Return the number points/vertices in [Polygon].
    */
   int get numberOfPoints => _List.length;
  
   /**
    * Return the point component of outer polygon.
    */
   GPoint getPoint(int index) {
      return _List[index];
   }
   
   List<GPoint> get points => _List;
   
   /**
    * Always [false] since [SimplePolygon]s cannot have holes.
    */
   bool get isHole => false;
   
   /**
    * Automatically throws [GeoClipException].
    * [SimplePolygon]s cannot be a holes.
    */
   void set isHole(bool isHole) {
      throw new GeoClipException("SimplePolygon cannot be a hole in and of itself.");
   }
   
   /**
    * [SimplePolygon]s can only have one polygon.
    *
    * @throws [GeoClipException] if polyIndex != 0
    */
   bool isContributing(int index) {
      if (index != 0) {
         throw new GeoClipException("SimplePolygon only has one polygon, itself.");
      }
      return _Contributes;
   }
   
   /**
    * Set whether or not this inside polygon is constributing to the set operation.
    *
    * @throws [GeoClipException] if polyIndex != 0
    */
   void setContributing(int index, bool contributes) {
      if (index != 0) {
         throw new GeoClipException("PolySimple only has one poly");
      }
      _Contributes = contributes;
   }
   
   /**
    * Return a [Polygon] that is the Intersection of this polygon.
    */
   SimplePolygon intersection(Polygon p) {
      return Boolean.intersection(this, p, "SimplePolygon");
   }
   
   /**
    * Return a [Polygon] that is the Union of this polygon.
    */
   SimplePolygon add(Polygon p) {
      return Boolean.add(this, p, "SimplePolygon");
   }
   
   /**
    * Return a [Polygon] that is the Exclusive-Or (xor) of this polygon.
    */
   SimplePolygon xor(Polygon p) {
      return Boolean.xor(p, this, "SimplePolygon");
   }
         
   /**
    * Return a [Polygon] that is the Difference of this polygon.
    */
   SimplePolygon subtract(Polygon p) {
      return Boolean.subtract(this, p, "SimplePolygon");
   }
         
   /**
    * Returns the Area of the [Polygon].
    */
  double get area {
    if (numberOfPoints < 3)
      return 0.0;
      
    GPoint a = getPoint(0);
    double area = 0.0 ;
    for( int i = 1 ; i < (numberOfPoints - 1) ; i++ ) {
      GPoint b = getPoint(i);
      GPoint c = getPoint(i+1);
      double tarea = ((c.x - b.x)*(a.y - b.y)) - ((a.x - b.x)*(c.y - b.y));
      area += tarea;
    }
    
    area = 0.5 * area.abs();
    
    return area ;
  }
}
