part of geoclip;

class _TopPolygonNode {
  _PolygonNode top_node = null;

  _PolygonNode add_local_min(double x, double y) {
    _PolygonNode existing_min = top_node;

    top_node = new _PolygonNode(existing_min, x, y);

    return top_node;
  }

  void merge_left(_PolygonNode p, _PolygonNode q) {
    // Label contour as a hole
    q.proxy.hole = true;

    if (p.proxy != q.proxy) {
      // Assign p's vertex list to the left end of q's list
      p.proxy.v[Boolean.RIGHT].next = q.proxy.v[Boolean.LEFT];
      q.proxy.v[Boolean.LEFT] = p.proxy.v[Boolean.LEFT];

      // Redirect any p.proxy references to q.proxy
      _PolygonNode target = p.proxy;
      for (_PolygonNode node = top_node; (node != null); node = node.next) {
        if (node.proxy == target) {
          node.active = 0;
          node.proxy = q.proxy;
        }
      }
    }
  }

  void merge_right(_PolygonNode p, _PolygonNode q) {
    // Label contour as external
    q.proxy.hole = false;

    if (p.proxy != q.proxy) {
      // Assign p's vertex list to the right end of q's list
      q.proxy.v[Boolean.RIGHT].next = p.proxy.v[Boolean.LEFT];
      q.proxy.v[Boolean.RIGHT] = p.proxy.v[Boolean.RIGHT];

      // Redirect any p->proxy references to q->proxy
      _PolygonNode target = p.proxy;
      for (_PolygonNode node = top_node; (node != null); node = node.next) {
        if (node.proxy == target) {
          node.active = 0;
          node.proxy = q.proxy;
        }
      }
    }
  }

  int count_contours() {
    int nc = 0;
    for (_PolygonNode polygon = top_node; (polygon != null); polygon =
        polygon.next) {
      if (polygon.active != 0) {
        // Count the vertices in the current contour
        int nv = 0;
        for (_VertexNode v = polygon.proxy.v[Boolean.LEFT]; (v != null); v =
            v.next) {
          nv++;
        }

        // Record valid vertex counts in the active field
        if (nv > 2) {
          polygon.active = nv;
          nc++;
        } else {
          // Invalid contour: just free the heap
          polygon.active = 0;
        }
      }
    }
    return nc;
  }

  Polygon getResult(String polyClass) {
    Polygon result = Boolean._createNewPolygon(polyClass);
    int num_contours = count_contours();
    if (num_contours > 0) {
      _PolygonNode npoly_node = null;
      for (_PolygonNode poly_node = top_node; (poly_node != null); poly_node =
          npoly_node) {
        npoly_node = poly_node.next;
        if (poly_node.active != 0) {
          Polygon poly = result;
          if (num_contours > 1) {
            poly = Boolean._createNewPolygon(polyClass);
          }
          if (poly_node.proxy.hole != null) {
            poly.isHole = poly_node.proxy.hole;
          }

          // This algorithm puts the verticies into the poly in reverse order
          for (_VertexNode vtx = poly_node.proxy.v[Boolean.LEFT]; (vtx != null);
              vtx = vtx.next) {
            poly.addCoord(vtx.x, vtx.y);
          }
          if (num_contours > 1) {
            result.addPolygon(poly);
          }
        }
      }

      // Sort holes to the end of the list
      Polygon orig = result;
      result = Boolean._createNewPolygon(polyClass);
      for (int i = 0; i < orig.numberOfPolygons; i++) {
        Polygon inner = orig.getPolygon(i);
        if (!inner.isHole) {
          result.addPolygon(inner);
        }
      }
      for (int i = 0; i < orig.numberOfPolygons; i++) {
        Polygon inner = orig.getPolygon(i);
        if (inner.isHole) {
          result.addPolygon(inner);
        }
      }
    }
    return result;
  }

  //      void print()
  //      {
  //         System.out.println("---- out_poly ----");
  //         int c= 0;
  //         _PolygonNode npoly_node = null ;
  //         for (_PolygonNode poly_node= top_node; (poly_node != null); poly_node = npoly_node)
  //         {
  //            System.out.println("contour="+c+"  active="+poly_node.active+"  hole="+poly_node.proxy.hole);
  //            npoly_node = poly_node.next;
  //            if (poly_node.active != 0)
  //            {
  //               int v=0 ;
  //               for (VertexNode vtx = poly_node.proxy.v[Boolean.LEFT]; (vtx != null) ; vtx = vtx.next )
  //               {
  //                  System.out.println("v="+v+"  vtx.x="+vtx.x+"  vtx.y="+vtx.y);
  //               }
  //               c++;
  //            }
  //         }
  //      }
}
