import 'dart:html';
import 'dart:math';
import 'dart:convert';

//import 'standard_tests.dart';
import 'package:geoclip/geoclip.dart' as Geo;
import 'shadows/shadow_map.dart';
import 'package:vector_math/vector_math.dart';

CanvasElement _canvas;

num _width;
num _height;
num _renderTime;
num _fpsAverage;
Element _fpsText = querySelector("#fps");
int _count = 0;
double _dirX = 0.55;

Element mousePosX = querySelector("#mousePosX");
Element mousePosY = querySelector("#mousePosY");

Geo.Polygon _subj = new Geo.ComplexPolygon();
Geo.Polygon _clip = new Geo.ComplexPolygon();

ShadowMap map;

void main() {
  // Uncomment the "import" and below to run the standard non-UI tests.
  //geoTests();
  //polygonTests();
  
  _canvas = querySelector("#area");
//  DivElement container = querySelector("#canvasContainer");
// 
//  canvas = new CanvasElement(width: viewSize.width, height: viewSize.height);
//  canvas.id = "gameCanvas";

  Rectangle rect = _canvas.parent.client;
  print("rect ${rect.width} x ${rect.height}");

  // Used for clearing rectangle.
  _width = rect.width;
  _height = rect.height;

  HttpRequest.getString("resources/app.json")
    .then(_processMap)
    .catchError(_handleLoadError);
}

void _processMap(String jsonString) {
  Map data = JSON.decode(jsonString);
  
  map = new ShadowMap(data, 1050.0, 1050.0);
  
//  _sutherlandHodgmanDataSet();
//  _clipperDataSet();

  _canvas.onMouseDown.listen(
      (event) => canvasMouseDown(event)
      );
  _canvas.onMouseMove.listen(
      (event) => canvasMouseMove(event)
      );
  _canvas.onMouseUp.listen(
      (event) => canvasMouseUp(event)
      );

//  _requestRedraw();
  draw(1);
}

void canvasMouseDown(MouseEvent event) {
  Point p = event.offset;
  map.touchDown(p);
  draw(1);
}

void canvasMouseUp(MouseEvent event) {
  Point p = event.offset;
  map.touchUp(p);
  draw(1);
}

void canvasMouseMove(MouseEvent event) {
  Point p = event.offset;
  map.touchMove(p);
  showMousePos(map.selectedLight.position);
  draw(1);
}

void _handleLoadError(Error error) {
  print("EmittersTab._handleDefaultSystemLoadError: $error");
}

void _clipperDataSet() {
  Geo.Polygon _square = new Geo.ComplexPolygon();
  _square.addCoord(180.0, 200.0);
  _square.addCoord(260.0, 200.0);  
  _square.addCoord(260.0, 150.0);
  _square.addCoord(180.0, 150.0);
  _subj.addPolygon(_square);

  Geo.Polygon _triangleHole = new Geo.ComplexPolygon.asHole(true);
  _triangleHole.addCoord(215.0, 160.0);
  _triangleHole.addCoord(230.0, 190.0);  
  _triangleHole.addCoord(200.0, 190.0);
  _subj.addPolygon(_triangleHole);

  _clip.addCoord(190.0, 210.0);
  _clip.addCoord(240.0, 210.0);  
  _clip.addCoord(240.0, 130.0);
  _clip.addCoord(190.0, 130.0);
}

void _sutherlandHodgmanDataSet() {
  _subj.addCoord(50.0, 150.0);
  _subj.addCoord(200.0, 50.0);
  _subj.addCoord(350.0, 150.0);
  _subj.addCoord(350.0, 300.0);
  _subj.addCoord(250.0, 300.0);
  _subj.addCoord(200.0, 250.0);
  _subj.addCoord(150.0, 350.0);
  _subj.addCoord(100.0, 250.0);
  _subj.addCoord(100.0, 200.0);

  double theta = 2.0 * PI;
  int segments = 16;
  double _theta_inc = 2.0 * PI / segments;
  double radius = 150.0;
  Geo.GPoint position = new Geo.GPoint(350.0, 200.0);

  // Note: Canvas +Y axis is oriented downwards.
  // So for CCW we decrement theta starting from 2PI.
  for(int i = 0; i < segments; i++) {
    double x = radius * cos(theta) + position.x;
    double y = radius * sin(theta) + position.y;
    _clip.addCoord(x, y);
    theta -= _theta_inc;
  }

//  _clip.addCoord(100.0, 100.0);
//  _clip.addCoord(300.0, 100.0);
//  _clip.addCoord(300.0, 300.0);
//  _clip.addCoord(100.0, 250.0);
}

void _requestRedraw() {
  window.requestAnimationFrame(draw);
}

void draw(num _) {
  num time = new DateTime.now().millisecondsSinceEpoch;
  if (_renderTime != null) showFps(1000 / (time - _renderTime));
  _renderTime = time;

  CanvasRenderingContext2D context = _canvas.context2D;
  drawBackground(context);
  
  _execute(context);
  
//  _requestRedraw();
}

void drawBackground(CanvasRenderingContext2D context) {
  context.clearRect(0, 0, _width, _height);
}

void showFps(num fps) {
  if (_fpsAverage == null) _fpsAverage = fps;
  _fpsAverage = fps * 0.05 + _fpsAverage * 0.95;
  _fpsText.text = "${_fpsAverage.round()} fps";
}

void showMousePos(Vector2 p) {
  mousePosX.text = "${p.x.toStringAsPrecision(8)}";
  mousePosY.text = "${p.y.toStringAsPrecision(8)}";
}

void _execute(CanvasRenderingContext2D context) {
  map.draw(context);
}

void _execute2(CanvasRenderingContext2D context) {
  if (_count > 300) {
    _dirX = -0.55;
  }
  else if (_count < -300) {
    _dirX = 0.55;
  }
  _count += _dirX.sign.toInt();

  _moveComplexPolygonByX(_clip, _dirX);
  
  drawComplexPolygonWithHoles(_subj, "0,0,255,0.082", "0,0,255,0.35", context);
  
  drawComplexPolygonWithHoles(_clip, "255,255,0,0.117", "255,0,0,0.176", context);
  
//  Geo.Polygon _solution = Geo.Boolean.intersectionComplex(_clip, _subj);
  Geo.Polygon _solution = Geo.Boolean.subtractComplex(_clip, _subj);
//  Geo.Polygon _solution = Geo.Boolean.addComplex(_clip, _subj);
//  Geo.Polygon _solution = Geo.Boolean.xorComplex(_clip, _subj);

  drawComplexPolygonWithHoles(_solution, "0,255,0,0.176", "0,90,0,1.0", context);
}

void _moveComplexPolygonByX(Geo.ComplexPolygon polygon, double dx) {
  List<Geo.Polygon> polygons = polygon.polygons;
  for(Geo.Polygon gp in polygons) {
    if (gp is Geo.ComplexPolygon)
      _moveComplexPolygonByX(gp, dx);
    else
      _moveSimplePolygonByX(gp, dx);
  }
}

void _moveSimplePolygonByX(Geo.SimplePolygon polygon, double dx) {
  List<Geo.GPoint> points = polygon.points;
  points.forEach((Geo.GPoint pt) => pt.x += dx);
}

void drawPolygon(Geo.Polygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
  if (polygon is Geo.SimplePolygon)
    drawSimplePolygon(polygon, fillColor, strokeColor, context);
  else
    drawComplexPolygon(polygon, fillColor, strokeColor, context);
}

void drawComplexPolygon(Geo.ComplexPolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
  // Complex polygons can have Simple polygons that are actually a part
  // of the Complex which makes them holes.
  // So we ask the Complex if it has any Simples acting as holes. If so
  // then we render differently.
  List<Geo.Polygon> polygons = polygon.polygons;
  
  if (polygon.hasHoles) {
    polygons.forEach((Geo.Polygon gp) => drawComplexPolygonWithHoles(gp, fillColor, strokeColor, context));
  }
  else {
    polygons.forEach((Geo.Polygon gp) => drawPolygon(gp, fillColor, strokeColor, context));
  }
  
}

void drawComplexPolygonWithHoles(Geo.ComplexPolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
  // This renderer only considers a simple case where SimplePolygons represent holes.
  // Other more complex arrangements will need a better renderer.
  if (polygon.isEmpty)
    return;

  context..lineWidth = 1
      ..strokeStyle = "rgba(" + strokeColor + ")"
      ..fillStyle = "rgba(" + fillColor + ")"
      ..beginPath();
  
  _drawComplexPolygonWithHoles(polygon, fillColor, strokeColor, context);
  
  context..fill()
      ..stroke();
}

void _drawComplexPolygonWithHoles(Geo.ComplexPolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
  for(Geo.Polygon ply in polygon.polygons) {
    if (ply is Geo.ComplexPolygon) {
      _drawComplexPolygonWithHoles(ply, fillColor, strokeColor, context);
    }
    else {
      Geo.SimplePolygon sp = ply as Geo.SimplePolygon;
      
      List<Geo.GPoint> points = sp.points;
      
      context.moveTo(points[0].x, points[0].y);
      
      Iterable<Geo.GPoint> subList = points.getRange(1, points.length);
      subList.forEach((Geo.GPoint p) => context.lineTo(p.x, p.y));
      
      context.closePath();
    }
  }
}

void drawSimplePolygon(Geo.SimplePolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
  if (polygon.isEmpty)
    return;
  
  List<Geo.GPoint> points = polygon.points;
  
  context..lineWidth = 1
      ..strokeStyle = "rgba(" + strokeColor + ")"
      ..fillStyle = "rgba(" + fillColor + ")"
      ..beginPath()
      ..moveTo(points[0].x, points[0].y);
      
  Iterable<Geo.GPoint> subList = points.getRange(1, points.length);
  subList.forEach((Geo.GPoint p) => context.lineTo(p.x, p.y));
//  for(Geo.GPoint p in subList) {
//    context.lineTo(p.x, p.y);
//  }
  context..closePath()
      ..fill()
      ..stroke();
}
