library caster;

import 'caster_edge.dart';
import 'package:vector_math/vector_math.dart';
import 'polygon.dart';

class Caster extends Polygon {
  List<CasterEdge> edgeList = new List<CasterEdge>();
  List<Vector2> wVertices = new List<Vector2>();

  Vector2 prev = new Vector2.zero();
  CasterEdge prevEdge = null;
  
  String normalColor = "white";
  
  bool showNormals = true;

  Map _map;
  
  Caster();

  void begin() {
    edgeList.clear();
  }

  void end() {
    Iterable<CasterEdge> subList = edgeList.sublist(1, edgeList.length);
    prevEdge = edgeList.first;
    prevEdge.calcAABBox();
    
    for (CasterEdge ce in subList) {
      ce.calcAABBox();
      prevEdge.nextEdge = ce;
      prevEdge = ce;
    }
    
    if (closed) {
      // Bind the first and last edges together.
      CasterEdge ce = edgeList.last;
      ce.calcAABBox();
      ce.nextEdge = edgeList.first;
      edgeList.first.prevEdge = ce;
    }
  }

  void startEdgeWith(double p0X, double p0Y, double p1X, double p1Y) {
    CasterEdge ce = new CasterEdge();
    ce.v0.setValues(p0X, p0Y);
    ce.v1.setValues(p1X, p1Y);
    ce.calcNormal();
    edgeList.add(ce);
    
    prev.setFrom(ce.v1);
    prevEdge = ce;
  }
  
  void appendEdgeWith(double pX, double pY) {
    CasterEdge ce = new CasterEdge();
    ce.v0.setFrom(prev);
    ce.v1.setValues(pX, pY);
    ce.calcNormal();
    edgeList.add(ce);
    
    prev.setFrom(ce.v1);

    ce.prevEdge = prevEdge;
    prevEdge = ce;
  }
  
  void close() {
    CasterEdge ce = new CasterEdge();
    ce.v0.setFrom(prev);
    
    CasterEdge ce0 = edgeList[0];
    ce.v1.setFrom(ce0.v0);
    ce.calcNormal();

    ce.prevEdge = prevEdge;
    edgeList.add(ce);
    closed = true;
  }
  
  Map toJson() {
    Map jCaster = new Map();

    jCaster["Position"] = {"x": position.x, "y": position.y};
    jCaster["Name"] = name;
    jCaster["Closed"] = closed;
    jCaster["Active"] = active;

    List<Map> jVertices = new List<Map>();
    
    vertices.forEach((Vector2 p) => jVertices.add({"x": p.x, "y": p.y}));
    
    jCaster["Vertices"] = jVertices;
    
    return jCaster;
  }

  Caster.fromJson(Map map) {
    _map = map;
    name = _map["Name"];
    active = _map["Active"] as bool;
    closed = _map["Closed"] as bool;
    
    num v = _map["Position"]["x"] as num;
    position.x = v.toDouble();
    v = _map["Position"]["y"] as num;
    position.y = v.toDouble();
    
    List<Map> jVertices = _map["Vertices"]; 
    begin();
    
    // These vertices list is in local-space
    // wVertices is world-space
    num sx = jVertices[0]["x"] as num;
    num sy = jVertices[0]["y"] as num;
    vertices.add(new Vector2(sx.toDouble(), sy.toDouble()));
    wVertices.add(new Vector2(sx.toDouble() + position.x, sy.toDouble() + position.y));
    
    num ex = jVertices[1]["x"] as num;
    num ey = jVertices[1]["y"] as num;
    vertices.add(new Vector2(ex.toDouble(), ey.toDouble()));
    wVertices.add(new Vector2(ex.toDouble() + position.x, ey.toDouble() + position.y));
    
    startEdgeWith(wVertices[0].x, wVertices[0].y, wVertices[1].x, wVertices[1].y);
    
    Iterable<Map> subList = jVertices.getRange(2, jVertices.length);
    
    for(Map p in subList) {
      num x = p["x"] as num;
      num y = p["y"] as num;
      vertices.add(new Vector2(x.toDouble(), y.toDouble()));
      wVertices.add(new Vector2(x.toDouble() + position.x, y.toDouble() + position.y));
      Vector2 last = wVertices.last;
      appendEdgeWith(last.x, last.y);
    }
    
    if (closed)
      close();
    
//    if (name == "Rock")
//      print(1);
    end();
    
    calcBounds(wVertices);
  }
  
  @override
  String toString() {
//    StringBuilder s = new StringBuilder();
//    s.append(name);
//    return s.toString();
    return name;
  }

}
