library caster_edge;

import 'package:vector_math/vector_math.dart';
import 'dart:math';

class CasterEdge {
  CasterEdge prevEdge = null;
  CasterEdge nextEdge = null;
  
  bool v0Visited = false;
  bool v1Visited = false;
  
  Vector2 v0 = new Vector2.zero();
  Vector2 v1 = new Vector2.zero();
  Vector2 _ray = new Vector2.zero();

  Vector2 _normal = new Vector2.zero();
  Vector2 _mid = new Vector2.zero();

  Aabb2 _aabbox;
  
  // deprecate
  // An edge can be a leaf=false or interior=true
  bool type = false;

  bool visited() {
    return v0Visited && v1Visited;
  }

  // Slightly faster with 1 less point allocation.
  void calcNormal() {
    // I want a CW normal which means x,y ==> -y,x.
    // Why? Because the coordinate system has +Y going downward in
    // screen-space.
    _normal = new Vector2(-(v1.y - v0.y), v1.x - v0.x);
    _normal = _normal.normalize();
  }
    
  Aabb2 get bounds => _aabbox;
    
  void calcAABBox() {
    double minX = min(v0.x, v1.x);
    double minY = min(v0.y, v1.y);
    double maxX = max(v0.x, v1.x);
    double maxY = max(v0.y, v1.y);
    
    _aabbox = new Aabb2.minMax(new Vector2(minX, minY), new Vector2(maxX, maxY));
  }
    
    /// Light's position is in shadow-space.
  bool isEdgeFacingLight(Vector2 lightPos) {
    // Form a ray that points towards the Light and compare
    // this with Edge's normal.
    
    // mid point
    _mid.x = (v0.x + v1.x) * 0.5;
    _mid.y = (v0.y + v1.y) * 0.5;

    Vector2 ray = new Vector2(_mid.x - lightPos.x, _mid.y - lightPos.y);
    ray = ray.normalize();
    
    // if dot == 1, they are facing exactly the same direction
    // if dot == -1, they are facing exactly the opposite direction
    // if dot > 0, they are facing roughly the same direction
    // if dot < 0, they are facing roughly the opposite direction
    // Because the Ray is non-normalized the dot product also
    // represents the distance of the Light from the edge.
    double dot = _normal.dot(ray);

    if (dot > 0.0)
      return false; // The edge is facing away from light. So the light is behind the edge.

    return true;
  }

  @override
  String toString() {
    StringBuffer s = new StringBuffer();
    s.write("v0:(${v0.x},${v0.y})");
    s.write("v1:(${v1.x},${v1.y})");
    s.write("n:(${_normal.x},${_normal.y})");
    return s.toString();
  }

}
