library caster_polygon;

import 'dart:math';
import 'dart:html';
import 'package:vector_math/vector_math.dart';

abstract class Polygon {
  // ----------------------------------------------------------
  // Vars
  // ----------------------------------------------------------
  static int _genId = 0;  // unique id.

  List<Vector2> vertices = new List<Vector2>();

  bool closed = false;

  String name = "";  

  bool active = false;

  bool dirty = true;
  Aabb2 bounds;
  
  Vector2 position = new Vector2.zero();

  double _prevX;
  double _prevY;
  Vector2 _prevVertex;

  String edgeColor = "rgba(255,0,255,1.0)";//"#724a16";
  String inActiveColor = "rgba(0,0,255,0.1)";

  Vector2 getVertex(int index) => vertices[index];

  void begin() {
    vertices.clear();
  }

  void end() {
  }

  void draw(CanvasRenderingContext2D context) {
    context.save();
    
    // First draw edges
    if (vertices.length > 0) {
      context.lineWidth = 1;
      
      if (active)
        context.strokeStyle = edgeColor;
      else
        context.strokeStyle = inActiveColor;

      Vector2 p = vertices[0];
      context..beginPath()
             ..moveTo(p.x + position.x, p.y + position.y);

      Iterable<Vector2> subList = vertices.getRange(1, vertices.length);
      for(Vector2 p in subList) {
        context.lineTo(p.x + position.x, p.y + position.y);
      }

      if (closed)
        context.closePath();

      context.stroke();
    }
    
//    for(Edge edge in edgeList) {
//      if (edge.culled) {
//        context.strokeStyle = edge.culledColor;
//        Vertex v0 = vertices[edge.p0];
//        Vertex v1 = vertices[edge.p1];
//        context..beginPath()
//               ..moveTo(v0.vertex.x, v0.vertex.y)
//               ..lineTo(v1.vertex.x, v1.vertex.y)
//               ..stroke();
//      }
//    }

    // Now draw normals.
//    if (showNormals) {
//      context.lineWidth = 2;
//
//      for(Edge e in edgeList) {
//        context.strokeStyle = e.normalColor;
//
//        Vector2 normal = e.normal;
//        Vector2 midPoint = e.mid;
//
//        context..beginPath()
//          ..moveTo(midPoint.x, midPoint.y)
//          ..lineTo(midPoint.x + (normal.x * 15.0), midPoint.y + (normal.y * 15.0))
//          ..stroke();
//      }
//    }
    
    context.restore();
  }

  void calcBounds(List<Vector2> vertices) {
    double minX = double.INFINITY;
    double minY = minX;
    double maxX = double.NEGATIVE_INFINITY;
    double maxY = maxX;
    
    for(Vector2 v in vertices) {
      minX = min(minX, v.x);
      minY = min(minY, v.y);
      maxX = max(maxX, v.x);
      maxY = max(maxY, v.y);
    }
    
    bounds = new Aabb2.minMax(new Vector2(minX, minY), new Vector2(maxX, maxY));
  }

  /// Returns a printable string
  @override
  String toString() {
    StringBuffer s = new StringBuffer();
    s.writeln("polygon '$name'");
    int i = 0;
    for(Vector2 vt in vertices) {
      s.writeln("  ${i} >> ${vt}");
      i++;
    }
    return s.toString();
  }

}