import 'package:geoclip/geoclip.dart' as Geo;

void geoTests() {
  testIntersectionEmptySet();
  testIntersectionOneContainsTwo();
  testIntersectionTwoContainsOne();
  testIntersectionTwoEqual();
  testIntersectionRectCorner1();
  testIntersectionRectCorner2();
  testIntersectionRectCorner3();
  testIntersectionRectCorner4();
  testIntersectionRectInterCorner1();
  testIntersectionRectInterCorner2();
  testIntersectionRectInterCorner3();
  testIntersectionRectInterCorner4();
  testIntersectionRectInterSide1();
  testIntersectionRectInterSide2();
  testIntersectionRectInterSide3();
  testIntersectionRectInterSide4();
  testIntersectionPolyOneOnTopOfTwo();
  testIntersectionPolyTwoSidesOneVertex();
  testIntersectionPolyTwoSides();
  testIntersectionPolyTwoSidesAndLowerVertex();
  testIntersectionPolyFourSides();
  testIntersectionPolyVOverlaps();
  testIntersectionRectangleHole();
  
  testUnionSeparate();
  testUnionOneContainsTwo();
  testUnionTwoContainsOne();

  testUnionTwoEqual();
  testUnionRectCorner1();
  testUnionRectCorner2();
  testUnionRectCorner3();
  testUnionRectCorner4();
  testUnionRectInterCorner1();
  testUnionRectInterCorner2();
  testUnionRectInterCorner3();
  testUnionRectInterCorner4();
  testUnionRectInterSide1();
  testUnionRectInterSide2();
  testUnionRectInterSide3();
  testUnionRectInterSide4();
  testUnionPolyOneOnTopOfTwo();
  testUnionPolyTwoSidesOneVertex();
  testUnionPolyTwoSides();
  testUnionPolyTwoSidesAndLowerVertex();
  testUnionPolyFourSides();
  testUnionPolyVOverlaps();
  testUnionRectangleHole();
  testXorSeparate();
  testXorOneContainsTwo();
  testXorTwoContainsOne();
  testXorTwoEqual();
  testXorRectCorner1();
  testXorRectCorner2();
  testXorRectCorner3();
  testXorRectCorner4();
  testXorRectInterCorner1();
  testXorRectInterCorner2();
  testXortRectInterCorner3();
  testXorRectInterCorner4();
  testXorRectInterSide1();
  testXornRectInterSide2();
  testXorRectInterSide3();
  testXorRectInterSide4();
  testXorPolyOneOnTopOfTwo();
  testXorPolyTwoSidesOneVertex();
  testXorPolyTwoSides();
  testXorPolyTwoSidesAndLowerVertex();
  testXorPolyFourSides();
  testXorPolyVOverlaps();
  testXorRectangleHole();
}

void polygonTests() {
  testEqualsEmpty();
  testEqualsSame();
  testEqualsDifferentOrder();
  testEqualsBadOrder();
  testEqualsDifferentNumPoints();
  testEqualsDifferentValue();
  testAreaTriangle();
  testAreaSquare();
  testAreaNonConvex();
  testAreaWithHole();
  testSelfIntersecting();
}

/**
 * Test the intersection of two polygons that are 
 * completely separate - result should be empty set.
 */
void testIntersectionEmptySet()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 1.0, 0.0 );
   p1.addCoord( 1.0, 1.0 );
   p1.addCoord( 0.0, 1.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 3.0 );
   
   Geo.Polygon empty = Geo.Boolean.intersectionComplex( p1, p2 );
   assert(empty.isEmpty);
   
   print("testIntersectionEmptySet complete");
}

/**
 * Test the intersection of two polygons where
 * the second is contained in the first.
 */
void testIntersectionOneContainsTwo()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 2.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.intersectionComplex( p1, p2 ) as Geo.ComplexPolygon;
   //print(result);
   assert( p2 == result );  // We are expecting equality here.
   print("testIntersectionOneContainsTwo complete");

}

/**
 * Test the intersection of two polygons where
 * the first is contained in the second.
 */
void testIntersectionTwoContainsOne()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 0.0 );
   p1.addCoord( 3.0, 0.0 );
   p1.addCoord( 3.0, 3.0 );
   p1.addCoord( 2.0, 3.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( p1 == result );
   print("testIntersectionTwoContainsOne complete");

}

/**
 * Test the intersection of two polygons that
 * are equal.
 */
void testIntersectionTwoEqual()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( p1 == result );
   print("testIntersectionTwoEqual complete");

}

/**
 * Test the intersection of two rectangles that share
 * one corner and two partial sides.
 */
void testIntersectionRectCorner1()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 1.0, 0.0 );
   p2.addCoord( 1.0, 1.0 );
   p2.addCoord( 0.0, 1.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( p2 == result );
   print("testIntersectionRectCorner1 complete");
}

/**
 * Test the intersection of two rectangles that share
 * one corner and two partial sides.
 */
void testIntersectionRectCorner2()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( p2 == result );
   print("testIntersectionRectCorner2 complete");
}

/**
 * Test the intersection of two rectangles that share
 * one corner and two partial sides.
 */
void testIntersectionRectCorner3()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 4.0, 3.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( p2 == result );
   print("testIntersectionRectCorner3 complete");
}

/**
 * Test the intersection of two rectangles that share
 * one corner and two partial sides.
 */
void testIntersectionRectCorner4()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 3.0 );
   p2.addCoord( 1.0, 3.0 );
   p2.addCoord( 1.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( p2 == result );
   print("testIntersectionRectCorner4 complete");
}

/**
 * Test the intersection of two rectangles that 
 * intersect on corner
 */
void testIntersectionRectInterCorner1()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 1.0, 3.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 2.0, 2.0 );
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 2.0, 3.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterCorner1 complete");
}

/**
 * Test the intersection of two rectangles that 
 * intersect on corner
 */
void testIntersectionRectInterCorner2()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 5.0, 1.0 );
   p2.addCoord( 5.0, 3.0 );
   p2.addCoord( 3.0, 3.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 4.0, 3.0 );
   exp.addCoord( 3.0, 3.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterCorner2 complete");
}

/**
 * Test the intersection of two rectangles that 
 * intersect on corner
 */
void testIntersectionRectInterCorner3()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 5.0, 3.0 );
   p2.addCoord( 5.0, 5.0 );
   p2.addCoord( 3.0, 5.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 4.0, 3.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 3.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterCorner3 complete");
}

/**
 * Test the intersection of two rectangles that 
 * intersect on corner
 */
void testIntersectionRectInterCorner4()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 3.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 3.0, 5.0 );
   p2.addCoord( 1.0, 5.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 2.0, 3.0 );
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 2.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterCorner4 complete");
}

/**
 * Test the intersection of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testIntersectionRectInterSide1()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 0.0 );
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 1.0, 2.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 1.0, 2.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterSide1 complete");
}

/**
 * Test the intersection of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testIntersectionRectInterSide2()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 5.0, 2.0 );
   p2.addCoord( 5.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 3.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterSide2 complete");
}

/**
 * Test the intersection of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testIntersectionRectInterSide3()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   p2.addCoord( 3.0, 6.0 );
   p2.addCoord( 1.0, 6.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 4.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 3.0, 5.0 );
   exp.addCoord( 1.0, 5.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterSide3 complete");
}

/**
 * Test the intersection of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testIntersectionRectInterSide4()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( -1.0, 2.0 );
   p2.addCoord(  1.0, 2.0 );
   p2.addCoord(  1.0, 4.0 );
   p2.addCoord( -1.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 2.0 );
   exp.addCoord( 1.0, 2.0 );
   exp.addCoord( 1.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionRectInterSide4 complete");
}

/**
 * Test the intersection of two complex, non-convex, non-self-intersecting
 * polygons - 1 on top of two = empty*/
void testIntersectionPolyOneOnTopOfTwo()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 4.0 );
   p1.addCoord( 5.0, 4.0 );
   p1.addCoord( 5.0, 9.0 );
   p1.addCoord( 3.0, 7.0 );
   p1.addCoord( 1.0, 9.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( result.isEmpty );
   print("testIntersectionPolyOneOnTopOfTwo complete");
}

/**
 * Test the intersection of two complex, non-convex, non-self-intersecting
 * polygons - two sides and one vertex - two triangles
 */
void testIntersectionPolyTwoSidesOneVertex()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 3.0 );
   p1.addCoord( 5.0, 3.0 );
   p1.addCoord( 5.0, 8.0 );
   p1.addCoord( 3.0, 6.0 );
   p1.addCoord( 1.0, 8.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );

   Geo.Polygon iexp1 = new Geo.ComplexPolygon();
   iexp1.addCoord( 4.0, 4.0 );
   iexp1.addCoord( 3.0, 3.0 );
   iexp1.addCoord( 5.0, 3.0 );
   Geo.Polygon iexp2 = new Geo.ComplexPolygon();
   iexp2.addCoord( 1.0, 3.0 );
   iexp2.addCoord( 3.0, 3.0 );
   iexp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( iexp1 );
   exp.addPolygon( iexp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.intersectionComplex( p1, p2 ) as Geo.ComplexPolygon;
//    print(result);
//    print(exp);
   assert( exp == result );
   print("testIntersectionPolyTwoSidesOneVertex complete");
}

/**
 * Test the intersection of two complex, non-convex, non-self-intersecting
 * polygons - two sides 
 */
void testIntersectionPolyTwoSides()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 2.0 );
   p1.addCoord( 5.0, 2.0 );
   p1.addCoord( 5.0, 7.0 );
   p1.addCoord( 3.0, 5.0 );
   p1.addCoord( 1.0, 7.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 2.0 );
   exp.addCoord( 5.0, 2.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.intersectionComplex( p1, p2 ) as Geo.ComplexPolygon;
   assert( exp == result );
   print("testIntersectionPolyTwoSides complete");
}

/**
 * Test the intersection of two complex, non-convex, non-self-intersecting
 * polygons - two sides and one vertex - the lower one
 */
void testIntersectionPolyTwoSidesAndLowerVertex()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 1.0 );
   p1.addCoord( 5.0, 1.0 );
   p1.addCoord( 5.0, 6.0 );
   p1.addCoord( 3.0, 4.0 );
   p1.addCoord( 1.0, 6.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 5.0, 1.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.intersectionComplex( p1, p2 ) as Geo.ComplexPolygon;
   assert( exp == result );
   print("testIntersectionPolyTwoSidesAndLowerVertex complete");

}

/**
 * Test the intersection of two complex, non-convex, non-self-intersecting
 * polygons - cross four sides
 */
void testIntersectionPolyFourSides()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 0.5 );
   p1.addCoord( 5.0, 0.5 );
   p1.addCoord( 5.0, 6.0 );
   p1.addCoord( 3.0, 4.0 );
   p1.addCoord( 1.0, 6.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 1.5, 0.5 );
   exp.addCoord( 2.5, 0.5 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 3.5, 0.5 );
   exp.addCoord( 4.5, 0.5 );
   exp.addCoord( 5.0, 1.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.Polygon result = Geo.Boolean.intersectionComplex( p1, p2 );
   assert( exp == result );
   print("testIntersectionPolyFourSides complete");
}

/**
 * Test the intersection of two complex, non-convex, non-self-intersecting
 * polygons - V overlap
 */
void testIntersectionPolyVOverlaps()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 0.0 );
   p1.addCoord( 5.0, 0.0 );
   p1.addCoord( 5.0, 5.0 );
   p1.addCoord( 3.0, 3.0 );
   p1.addCoord( 1.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 2.0, 0.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 5.0, 1.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.intersectionComplex( p1, p2 ) as Geo.ComplexPolygon;
   assert( exp == result );
   print("testIntersectionPolyVOverlaps complete");
}

/**
 * Test the intersection of a rectangle with a hole and solid rectangle
 */
void testIntersectionRectangleHole()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon.asHole(true);
   p2.addCoord( 1.0, 2.0 );
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 3.0, 4.0 );
   p2.addCoord( 1.0, 4.0 );
   
   Geo.Polygon p12 = new Geo.ComplexPolygon();
   p12.addPolygon( p1 );
   p12.addPolygon( p2 );

   Geo.Polygon p3 = new Geo.ComplexPolygon();
   p3.addCoord( 2.0, 0.0 );
   p3.addCoord( 6.0, 0.0 );
   p3.addCoord( 6.0, 6.0 );
   p3.addCoord( 2.0, 6.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 2.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 2.0, 5.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 2.0, 2.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.intersectionComplex( p12, p3 ) as Geo.ComplexPolygon;
   assert( exp == result );
   print("testIntersectionRectangleHole complete");
}

/**
 * Test the UNION of two polygons that are 
 * completely separate - result should be poly that contains these two polys
 */
void testUnionSeparate()
{
  Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 1.0, 0.0 );
   p1.addCoord( 1.0, 1.0 );
   p1.addCoord( 0.0, 1.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 3.0 );

   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( p2 );
   exp.addPolygon( p1 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 ) as Geo.ComplexPolygon;
   assert( exp == result );
   print("testUnionSeparate complete");
}

/**
 * Test the union of two polygons where
 * the second is contained in the first.
 */
void testUnionOneContainsTwo()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 2.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
//      result.print();
   assert( p1 == result );
   print("testUnionOneContainsTwo complete");
}
/**
 * Test the union of two polygons where
 * the first is contained in the second.
 */
void testUnionTwoContainsOne()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 0.0 );
   p1.addCoord( 3.0, 0.0 );
   p1.addCoord( 3.0, 3.0 );
   p1.addCoord( 2.0, 3.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
//      result.print();
   assert( p2 == result );
   print("testUnionTwoContainsOne complete");
}

/**
 * Test the union of two polygons that
 * are equal.
 */
void testUnionTwoEqual()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
//      result.print();
   assert( p1 == result );
   print("testUnionTwoEqual complete");
}

/**
 * Test the union of two rectangles that share
 * one corner and two partial sides.
 */
void testUnionRectCorner1()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 1.0, 0.0 );
   p2.addCoord( 1.0, 1.0 );
   p2.addCoord( 0.0, 1.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 0.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   exp.addCoord( 0.0, 1.0 ); // !!! KNOWN BUG - EXTRA POINT BUT SHAPE IS CORRECT
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectCorner1 complete");
}

/**
 * Test the UNION of two rectangles that share
 * one corner and two partial sides.
 */
void testUnionRectCorner2()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 0.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.0, 1.0 ); // !!! KNOWN BUG - EXTRA POINT BUT SHAPE IS CORRECT
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectCorner2 complete");
}

/**
 * Test the union of two rectangles that share
 * one corner and two partial sides.
 */
void testUnionRectCorner3()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 4.0, 3.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 0.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.0, 3.0 ); // !!! KNOWN BUG - EXTRA POINT BUT SHAPE IS CORRECT
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectCorner3 complete");
}

/**
 * Test the union of two rectangles that share
 * one corner and two partial sides.
 */
void testUnionRectCorner4()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 3.0 );
   p2.addCoord( 1.0, 3.0 );
   p2.addCoord( 1.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 0.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   exp.addCoord( 0.0, 3.0 ); // !!! KNOWN BUG - EXTRA POINT BUT SHAPE IS CORRECT
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectCorner4 complete");
}

/**
 * Test the union of two rectangles that 
 * intersect on corner
 */
void testUnionRectInterCorner1()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 1.0, 3.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 2.0, 3.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterCorner1 complete");
}

/**
 * Test the union of two rectangles that 
 * intersect on corner
 */
void testUnionRectInterCorner2()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 5.0, 1.0 );
   p2.addCoord( 5.0, 3.0 );
   p2.addCoord( 3.0, 3.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 5.0, 1.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 4.0, 3.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 2.0, 2.0 );
   exp.addCoord( 3.0, 2.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterCorner2 complete");
}
   
/**
 * Test the union of two rectangles that 
 * intersect on corner
 */
void testUnionRectInterCorner3()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 5.0, 3.0 );
   p2.addCoord( 5.0, 5.0 );
   p2.addCoord( 3.0, 5.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 5.0, 5.0 );
   exp.addCoord( 3.0, 5.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 2.0, 2.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 4.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterCorner3 complete");
}

/**
 * Test the union of two rectangles that 
 * intersect on corner
 */
void testUnionRectInterCorner4()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 3.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 3.0, 5.0 );
   p2.addCoord( 1.0, 5.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 3.0 );
   exp.addCoord( 2.0, 3.0 );
   exp.addCoord( 2.0, 2.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 3.0, 5.0 );
   exp.addCoord( 1.0, 5.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterCorner4 complete");
}
   
/**
 * Test the union of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testUnionRectInterSide1()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 0.0 );
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 1.0, 2.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 0.0 );
   exp.addCoord( 3.0, 0.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 0.0, 5.0 );
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 1.0, 1.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterSide1 complete");
}

/**
 * Test the union of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testUnionRectInterSide2()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 5.0, 2.0 );
   p2.addCoord( 5.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 5.0, 2.0 );
   exp.addCoord( 5.0, 4.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 0.0, 5.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterSide2 complete");
}
   
/**
 * Test the union of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testUnionRectInterSide3()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   p2.addCoord( 3.0, 6.0 );
   p2.addCoord( 1.0, 6.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 3.0, 5.0 );
   exp.addCoord( 3.0, 6.0 );
   exp.addCoord( 1.0, 6.0 );
   exp.addCoord( 1.0, 5.0 );
   exp.addCoord( 0.0, 5.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterSide3 complete");
}

/**
 * Test the union of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testUnionRectInterSide4()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( -1.0, 2.0 );
   p2.addCoord(  1.0, 2.0 );
   p2.addCoord(  1.0, 4.0 );
   p2.addCoord( -1.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord(  0.0, 1.0 );
   exp.addCoord(  4.0, 1.0 );
   exp.addCoord(  4.0, 5.0 );
   exp.addCoord(  0.0, 5.0 );
   exp.addCoord(  0.0, 4.0 );
   exp.addCoord( -1.0, 4.0 );
   exp.addCoord( -1.0, 2.0 );
   exp.addCoord(  0.0, 2.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionRectInterSide4 complete");
}

/**
 * Test the union of two complex, non-convex, non-self-intersecting
 * polygons - 1 on top of two */
void testUnionPolyOneOnTopOfTwo()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 4.0 );
   p1.addCoord( 5.0, 4.0 );
   p1.addCoord( 5.0, 9.0 );
   p1.addCoord( 3.0, 7.0 );
   p1.addCoord( 1.0, 9.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 0.0, 2.0 );
   exp1.addCoord( 2.0, 0.0 );
   exp1.addCoord( 3.0, 1.0 );
   exp1.addCoord( 4.0, 0.0 );
   exp1.addCoord( 6.0, 2.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 5.0, 4.0 );
   exp1.addCoord( 5.0, 9.0 );
   exp1.addCoord( 3.0, 7.0 );
   exp1.addCoord( 1.0, 9.0 );
   exp1.addCoord( 1.0, 4.0 );
   exp1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon.asHole(true);
   exp2.addCoord( 4.0, 4.0 );
   exp2.addCoord( 3.0, 3.0 );
   exp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionPolyOneOnTopOfTwo complete");
}

/**
 * Test the union of two complex, non-convex, non-self-intersecting
 * polygons - two sides and one vertex - two triangles
 */
void testUnionPolyTwoSidesOneVertex()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 3.0 );
   p1.addCoord( 5.0, 3.0 );
   p1.addCoord( 5.0, 8.0 );
   p1.addCoord( 3.0, 6.0 );
   p1.addCoord( 1.0, 8.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );

   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 2.0 );
   exp.addCoord( 2.0, 0.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 6.0, 2.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 5.0, 8.0 );
   exp.addCoord( 3.0, 6.0 );
   exp.addCoord( 1.0, 8.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionPolyTwoSidesOneVertex complete");
}

/**
 * Test the union of two complex, non-convex, non-self-intersecting
 * polygons - two sides 
 */
void testUnionPolyTwoSides()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 2.0 );
   p1.addCoord( 5.0, 2.0 );
   p1.addCoord( 5.0, 7.0 );
   p1.addCoord( 3.0, 5.0 );
   p1.addCoord( 1.0, 7.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 2.0 );
   exp.addCoord( 2.0, 0.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 6.0, 2.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 5.0, 7.0 );
   exp.addCoord( 3.0, 5.0 );
   exp.addCoord( 1.0, 7.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionPolyTwoSides complete");
}

/**
 * Test the union of two complex, non-convex, non-self-intersecting
 * polygons - two sides and one vertex - the lower one
 */
void testUnionPolyTwoSidesAndLowerVertex()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 1.0 );
   p1.addCoord( 5.0, 1.0 );
   p1.addCoord( 5.0, 6.0 );
   p1.addCoord( 3.0, 4.0 );
   p1.addCoord( 1.0, 6.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 2.0 );
   exp.addCoord( 2.0, 0.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 6.0, 2.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 5.0, 6.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 1.0, 6.0 );
   exp.addCoord( 1.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionPolyTwoSidesAndLowerVertex complete");
}

/**
 * Test the union of two complex, non-convex, non-self-intersecting
 * polygons - cross four sides
 */
void testUnionPolyFourSides()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 0.5 );
   p1.addCoord( 5.0, 0.5 );
   p1.addCoord( 5.0, 6.0 );
   p1.addCoord( 3.0, 4.0 );
   p1.addCoord( 1.0, 6.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 1.0, 0.5 );
   exp.addCoord( 1.5, 0.5 );
   exp.addCoord( 2.0, 0.0 );
   exp.addCoord( 2.5, 0.5 );
   exp.addCoord( 3.5, 0.5 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.5, 0.5 );
   exp.addCoord( 5.0, 0.5 );
   exp.addCoord( 5.0, 1.0 );
   exp.addCoord( 6.0, 2.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 5.0, 6.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 1.0, 6.0 );
   exp.addCoord( 1.0, 3.0 );
   exp.addCoord( 0.0, 2.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionPolyFourSides complete");
}
      
/**
 * Test the union of two complex, non-convex, non-self-intersecting
 * polygons - V overlap
 */
void testUnionPolyVOverlaps()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 0.0 );
   p1.addCoord( 5.0, 0.0 );
   p1.addCoord( 5.0, 5.0 );
   p1.addCoord( 3.0, 3.0 );
   p1.addCoord( 1.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 0.0 );
   exp.addCoord( 5.0, 0.0 );
   exp.addCoord( 5.0, 1.0 );
   exp.addCoord( 6.0, 2.0 );
   exp.addCoord( 5.0, 3.0 );
   exp.addCoord( 5.0, 5.0 );
   exp.addCoord( 4.0, 4.0 ); // KNOWN BUG - EXTRA POINT BUT SHAPE IS OK
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 2.0, 4.0 ); // KNOWN BUG - EXTRA POINT BUT SHAPE IS OK
   exp.addCoord( 1.0, 5.0 );
   exp.addCoord( 1.0, 3.0 );
   exp.addCoord( 0.0, 2.0 );
   exp.addCoord( 1.0, 1.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testUnionPolyVOverlaps complete");
}

/**
 * Test the UNION of a rectangle with a hole and solid rectangle
 */
void testUnionRectangleHole()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon.asHole(true);
   p2.addCoord( 1.0, 2.0 );
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 3.0, 4.0 );
   p2.addCoord( 1.0, 4.0 );
   
   Geo.Polygon p12 = new Geo.ComplexPolygon();
   p12.addPolygon( p1 );
   p12.addPolygon( p2 );

   Geo.Polygon p3 = new Geo.ComplexPolygon();
   p3.addCoord( 2.0, 0.0 );
   p3.addCoord( 6.0, 0.0 );
   p3.addCoord( 6.0, 6.0 );
   p3.addCoord( 2.0, 6.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 2.0, 0.0 );
   exp1.addCoord( 6.0, 0.0 );
   exp1.addCoord( 6.0, 6.0 );
   exp1.addCoord( 2.0, 6.0 );
   exp1.addCoord( 2.0, 5.0 );
   exp1.addCoord( 0.0, 5.0 );
   exp1.addCoord( 0.0, 1.0 );
   exp1.addCoord( 2.0, 1.0 );

   Geo.Polygon exp2 = new Geo.ComplexPolygon.asHole(true);
   exp2.addCoord( 2.0, 2.0 );
   exp2.addCoord( 1.0, 2.0 );
   exp2.addCoord( 1.0, 4.0 );
   exp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.addComplex( p12, p3 );
   //result.print();
   assert( exp == result );
   print("testUnionRectangleHole complete");
}

// -----------
// --- XOR ---
// -----------
/**
 * Test the XOR of two polygons that are 
 * completely separate - result should be poly that contains these two polys
 */
void testXorSeparate()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 1.0, 0.0 );
   p1.addCoord( 1.0, 1.0 );
   p1.addCoord( 0.0, 1.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 3.0 );

   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( p2 );
   exp.addPolygon( p1 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
//      result.print();
   assert( exp == result );
   print("testXorSeparate complete");
}

/**
 * Test the xor of two polygons where
 * the second is contained in the first.
 */
void testXorOneContainsTwo()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 2.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 3.0 );
   
   // notice reverse order
   Geo.Polygon exp2 = new Geo.ComplexPolygon.asHole(true);
   exp2.addCoord( 3.0, 1.0 );
   exp2.addCoord( 2.0, 1.0 );
   exp2.addCoord( 2.0, 3.0 );
   exp2.addCoord( 3.0, 3.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( p1 );
   exp.addPolygon( exp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorOneContainsTwo complete");
}

/**
 * Test the xor of two polygons where
 * the first is contained in the second.
 */
void testXorTwoContainsOne()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 1.0 );
   p1.addCoord( 3.0, 1.0 );
   p1.addCoord( 3.0, 3.0 );
   p1.addCoord( 2.0, 3.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   // notice reverse order
   Geo.Polygon exp2 = new Geo.ComplexPolygon.asHole(true);
   exp2.addCoord( 3.0, 1.0 );
   exp2.addCoord( 2.0, 1.0 );
   exp2.addCoord( 2.0, 3.0 );
   exp2.addCoord( 3.0, 3.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( p2 );
   exp.addPolygon( exp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorTwoContainsOne complete");
}

/**
 * Test the xor of two polygons that
 * are equal.
 */
void testXorTwoEqual()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( result.isEmpty );
   print("testXorTwoEqual complete");
}

/**
 * Test the xor of two rectangles that share
 * one corner and two partial sides.
 */
void testXorRectCorner1()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 0.0 );
   p2.addCoord( 1.0, 0.0 );
   p2.addCoord( 1.0, 1.0 );
   p2.addCoord( 0.0, 1.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 1.0, 0.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 1.0, 1.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectCorner1 complete");
}
   
/**
 * Test the xor of two rectangles that share
 * one corner and two partial sides.
 */
void testXorRectCorner2()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 4.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 0.0 );
   exp.addCoord( 3.0, 0.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectCorner2 complete");
}

/**
 * Test the xor of two rectangles that share
 * one corner and two partial sides.
 */
void testXorRectCorner3()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 4.0, 3.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 0.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.0, 3.0 );
   exp.addCoord( 3.0, 3.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 0.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectCorner3 complete");
}

/**
 * Test the xor of two rectangles that share
 * one corner and two partial sides.
 */
void testXorRectCorner4()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 0.0 );
   p1.addCoord( 4.0, 0.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 0.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 3.0 );
   p2.addCoord( 1.0, 3.0 );
   p2.addCoord( 1.0, 4.0 );
   p2.addCoord( 0.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 0.0 );
   exp.addCoord( 4.0, 0.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 1.0, 4.0 );
   exp.addCoord( 1.0, 3.0 );
   exp.addCoord( 0.0, 3.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectCorner4 complete");
}   

/**
 * Test the xor of two rectangles that 
 * intersect on corner
 */
void testXorRectInterCorner1()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 1.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 1.0, 3.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 3.0, 2.0 );
   exp1.addCoord( 4.0, 2.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 2.0, 4.0 );
   exp1.addCoord( 2.0, 3.0 );
   exp1.addCoord( 3.0, 3.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 1.0, 1.0 );
   exp2.addCoord( 3.0, 1.0 );
   exp2.addCoord( 3.0, 2.0 );
   exp2.addCoord( 2.0, 2.0 );
   exp2.addCoord( 2.0, 3.0 );
   exp2.addCoord( 1.0, 3.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectInterCorner1 complete");
}

/**
 * Test the xor of two rectangles that 
 * intersect on corner
 */
void testXorRectInterCorner2()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 5.0, 1.0 );
   p2.addCoord( 5.0, 3.0 );
   p2.addCoord( 3.0, 3.0 );
   
   // ------------------------------------------------------------------------------------------
   // --- I expected this to give two non-hole inner polygons but it gave one with a hole    ---
   // --- if you look at it, they are equivalent.  Don't have time to figure out difference. ---
   // ------------------------------------------------------------------------------------------
//      Geo.Polygon exp1 = new Geo.ComplexPolygon();
//      exp1.addCoord( 3.0, 1.0 );
//      exp1.addCoord( 5.0, 1.0 );
//      exp1.addCoord( 5.0, 3.0 );
//      exp1.addCoord( 4.0, 3.0 );
//      exp1.addCoord( 4.0, 2.0 );
//      exp1.addCoord( 3.0, 2.0 );
//      
//      Geo.Polygon exp2 = new Geo.ComplexPolygon();
//      exp2.addCoord( 2.0, 2.0 );
//      exp2.addCoord( 3.0, 2.0 );
//      exp2.addCoord( 3.0, 3.0 );
//      exp2.addCoord( 4.0, 3.0 );
//      exp2.addCoord( 4.0, 4.0 );
//      exp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon.asHole(true);
   exp1.addCoord( 3.0, 2.0 );
   exp1.addCoord( 3.0, 3.0 );
   exp1.addCoord( 4.0, 3.0 );
   exp1.addCoord( 4.0, 2.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 2.0, 2.0 );
   exp2.addCoord( 3.0, 2.0 );
   exp2.addCoord( 3.0, 1.0 );
   exp2.addCoord( 5.0, 1.0 );
   exp2.addCoord( 5.0, 3.0 );
   exp2.addCoord( 4.0, 3.0 );
   exp2.addCoord( 4.0, 4.0 );
   exp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp2 );
   exp.addPolygon( exp1 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectInterCorner2 complete");
}
   

/**
 * Test the xor of two rectangles that 
 * intersect on corner
 */
void testXortRectInterCorner3()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 5.0, 3.0 );
   p2.addCoord( 5.0, 5.0 );
   p2.addCoord( 3.0, 5.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 2.0, 2.0 );
   exp1.addCoord( 4.0, 2.0 );
   exp1.addCoord( 4.0, 3.0 );
   exp1.addCoord( 3.0, 3.0 );
   exp1.addCoord( 3.0, 4.0 );
   exp1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 4.0, 3.0 );
   exp2.addCoord( 5.0, 3.0 );
   exp2.addCoord( 5.0, 5.0 );
   exp2.addCoord( 3.0, 5.0 );
   exp2.addCoord( 3.0, 4.0 );
   exp2.addCoord( 4.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp2 );
   exp.addPolygon( exp1 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXortRectInterCorner3 complete");
}

/**
 * Test the xor of two rectangles that 
 * intersect on corner
 */
void testXorRectInterCorner4()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 2.0, 2.0 );
   p1.addCoord( 4.0, 2.0 );
   p1.addCoord( 4.0, 4.0 );
   p1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 3.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 3.0, 5.0 );
   p2.addCoord( 1.0, 5.0 );
   
   // ------------------------------------------------------------------------------------------
   // --- I expected this to give two non-hole inner polygons but it gave one with a hole    ---
   // --- if you look at it, they are equivalent.  Don't have time to figure out difference. ---
   // ------------------------------------------------------------------------------------------

   Geo.Polygon exp1 = new Geo.ComplexPolygon.asHole(true);
   exp1.addCoord( 3.0, 3.0 );
   exp1.addCoord( 2.0, 3.0 );
   exp1.addCoord( 2.0, 4.0 );
   exp1.addCoord( 3.0, 4.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 1.0, 3.0 );
   exp2.addCoord( 2.0, 3.0 );
   exp2.addCoord( 2.0, 2.0 );
   exp2.addCoord( 4.0, 2.0 );
   exp2.addCoord( 4.0, 4.0 );
   exp2.addCoord( 3.0, 4.0 );
   exp2.addCoord( 3.0, 5.0 );
   exp2.addCoord( 1.0, 5.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp2 );
   exp.addPolygon( exp1 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectInterCorner4 complete");
}

/**
 * Test the xor of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testXorRectInterSide1()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 0.0 );
   p2.addCoord( 3.0, 0.0 );
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 1.0, 2.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 1.0, 0.0 );
   exp.addCoord( 3.0, 0.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 1.0, 1.0 );
   exp.addCoord( 1.0, 2.0 );
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 3.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 0.0, 5.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectInterSide1 complete");
}
   
/**
 * Test the xor of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testXornRectInterSide2()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 5.0, 2.0 );
   p2.addCoord( 5.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 0.0, 5.0 );
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 4.0, 4.0 );
   exp.addCoord( 4.0, 2.0 );
   exp.addCoord( 5.0, 2.0 );
   exp.addCoord( 5.0, 4.0 );
   exp.addCoord( 4.0, 4.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXornRectInterSide2 complete");
}

/**
 * Test the xor of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testXorRectInterSide3()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 1.0, 4.0 );
   p2.addCoord( 3.0, 4.0 );
   p2.addCoord( 3.0, 6.0 );
   p2.addCoord( 1.0, 6.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 3.0, 5.0 );
   exp.addCoord( 3.0, 6.0 );
   exp.addCoord( 1.0, 6.0 );
   exp.addCoord( 1.0, 5.0 );
   exp.addCoord( 3.0, 5.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 1.0, 4.0 );
   exp.addCoord( 1.0, 5.0 );
   exp.addCoord( 0.0, 5.0 );
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 4.0, 1.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectInterSide3 complete");
}
   
/**
 * Test the xor of two rectangles where
 * half of one is contained in the other and
 * two sides of the inner cross one side of the outer.
 */
void testXorRectInterSide4()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( -1.0, 2.0 );
   p2.addCoord(  1.0, 2.0 );
   p2.addCoord(  1.0, 4.0 );
   p2.addCoord( -1.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord(  4.0, 5.0 );
   exp.addCoord(  0.0, 5.0 );
   exp.addCoord(  0.0, 4.0 );
   exp.addCoord(  1.0, 4.0 );
   exp.addCoord(  1.0, 2.0 );
   exp.addCoord(  0.0, 2.0 );
   exp.addCoord(  0.0, 4.0 );
   exp.addCoord( -1.0, 4.0 );
   exp.addCoord( -1.0, 2.0 );
   exp.addCoord(  0.0, 2.0 );
   exp.addCoord(  0.0, 1.0 );
   exp.addCoord(  4.0, 1.0 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorRectInterSide4 complete");
}
   
/**
 * Test the xor of two complex, non-convex, non-self-intersecting
 * polygons - 1 on top of two*/
void testXorPolyOneOnTopOfTwo()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 4.0 );
   p1.addCoord( 5.0, 4.0 );
   p1.addCoord( 5.0, 9.0 );
   p1.addCoord( 3.0, 7.0 );
   p1.addCoord( 1.0, 9.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 0.0, 2.0 );
   exp1.addCoord( 2.0, 0.0 );
   exp1.addCoord( 3.0, 1.0 );
   exp1.addCoord( 4.0, 0.0 );
   exp1.addCoord( 6.0, 2.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 5.0, 4.0 );
   exp1.addCoord( 5.0, 9.0 );
   exp1.addCoord( 3.0, 7.0 );
   exp1.addCoord( 1.0, 9.0 );
   exp1.addCoord( 1.0, 4.0 );
   exp1.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon.asHole(true);
   exp2.addCoord( 4.0, 4.0 );
   exp2.addCoord( 3.0, 3.0 );
   exp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorPolyOneOnTopOfTwo complete");
}

/**
 * Test the xor of two complex, non-convex, non-self-intersecting
 * polygons - two sides and one vertex - two triangles
 */
void testXorPolyTwoSidesOneVertex()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 3.0 );
   p1.addCoord( 5.0, 3.0 );
   p1.addCoord( 5.0, 8.0 );
   p1.addCoord( 3.0, 6.0 );
   p1.addCoord( 1.0, 8.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );

   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 5.0, 8.0 );
   exp1.addCoord( 3.0, 6.0 );
   exp1.addCoord( 1.0, 8.0 );
   exp1.addCoord( 1.0, 3.0 );
   exp1.addCoord( 0.0, 2.0 );
   exp1.addCoord( 2.0, 0.0 );
   exp1.addCoord( 3.0, 1.0 );
   exp1.addCoord( 4.0, 0.0 );
   exp1.addCoord( 6.0, 2.0 );
   exp1.addCoord( 5.0, 3.0 );
   exp1.addCoord( 3.0, 3.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 5.0, 3.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon.asHole(true);
   exp2.addCoord( 3.0, 3.0 );
   exp2.addCoord( 1.0, 3.0 );
   exp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorPolyTwoSidesOneVertex complete");
}
   
/**
 * Test the xor of two complex, non-convex, non-self-intersecting
 * polygons - two sides 
 */
void testXorPolyTwoSides()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 2.0 );
   p1.addCoord( 5.0, 2.0 );
   p1.addCoord( 5.0, 7.0 );
   p1.addCoord( 3.0, 5.0 );
   p1.addCoord( 1.0, 7.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 5.0, 7.0 );
   exp1.addCoord( 3.0, 5.0 );
   exp1.addCoord( 1.0, 7.0 );
   exp1.addCoord( 1.0, 3.0 );
   exp1.addCoord( 2.0, 4.0 );
   exp1.addCoord( 3.0, 3.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 5.0, 3.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 5.0, 3.0 );
   exp2.addCoord( 5.0, 2.0 );
   exp2.addCoord( 1.0, 2.0 );
   exp2.addCoord( 1.0, 3.0 );
   exp2.addCoord( 0.0, 2.0 );
   exp2.addCoord( 2.0, 0.0 );
   exp2.addCoord( 3.0, 1.0 );
   exp2.addCoord( 4.0, 0.0 );
   exp2.addCoord( 6.0, 2.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );      
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorPolyTwoSides complete");
}
   
/**
 * Test the xor of two complex, non-convex, non-self-intersecting
 * polygons - two sides and one vertex - the lower one
 */
void testXorPolyTwoSidesAndLowerVertex()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 1.0 );
   p1.addCoord( 5.0, 1.0 );
   p1.addCoord( 5.0, 6.0 );
   p1.addCoord( 3.0, 4.0 );
   p1.addCoord( 1.0, 6.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 5.0, 6.0 );
   exp1.addCoord( 3.0, 4.0 );
   exp1.addCoord( 1.0, 6.0 );
   exp1.addCoord( 1.0, 3.0 );
   exp1.addCoord( 2.0, 4.0 );
   exp1.addCoord( 3.0, 3.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 5.0, 3.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 5.0, 3.0 );
   exp2.addCoord( 5.0, 1.0 );
   exp2.addCoord( 3.0, 1.0 );
   exp2.addCoord( 4.0, 0.0 );
   exp2.addCoord( 6.0, 2.0 );
   
   Geo.Polygon exp3 = new Geo.ComplexPolygon();
   exp3.addCoord( 1.0, 3.0 );
   exp3.addCoord( 0.0, 2.0 );
   exp3.addCoord( 2.0, 0.0 );
   exp3.addCoord( 3.0, 1.0 );
   exp3.addCoord( 1.0, 1.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   exp.addPolygon( exp3 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorPolyTwoSidesAndLowerVertex complete");
}

/**
 * Test the xor of two complex, non-convex, non-self-intersecting
 * polygons - cross four sides
 */
void testXorPolyFourSides()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 0.5 );
   p1.addCoord( 5.0, 0.5 );
   p1.addCoord( 5.0, 6.0 );
   p1.addCoord( 3.0, 4.0 );
   p1.addCoord( 1.0, 6.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 5.0, 6.0 );
   exp1.addCoord( 3.0, 4.0 );
   exp1.addCoord( 1.0, 6.0 );
   exp1.addCoord( 1.0, 3.0 );
   exp1.addCoord( 2.0, 4.0 );
   exp1.addCoord( 3.0, 3.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 5.0, 3.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 5.0, 3.0 );
   exp2.addCoord( 5.0, 1.0 );
   exp2.addCoord( 6.0, 2.0 );
   
   Geo.Polygon exp3 = new Geo.ComplexPolygon();
   exp3.addCoord( 1.0, 3.0 );
   exp3.addCoord( 0.0, 2.0 );
   exp3.addCoord( 1.0, 1.0 );
   
   Geo.Polygon exp4 = new Geo.ComplexPolygon();
   exp4.addCoord( 5.0, 1.0 );
   exp4.addCoord( 4.5, 0.5 );
   exp4.addCoord( 5.0, 0.5 );
   
   Geo.Polygon exp5 = new Geo.ComplexPolygon();
   exp5.addCoord( 3.0, 1.0 );
   exp5.addCoord( 2.5, 0.5 );
   exp5.addCoord( 3.5, 0.5 );
   exp5.addCoord( 4.0, 0.0 );
   exp5.addCoord( 4.5, 0.5 );
   exp5.addCoord( 3.5, 0.5 );
   
   Geo.Polygon exp6 = new Geo.ComplexPolygon();
   exp6.addCoord( 1.0, 1.0 );
   exp6.addCoord( 1.0, 0.5 );
   exp6.addCoord( 1.5, 0.5 );
   exp6.addCoord( 2.0, 0.0 );
   exp6.addCoord( 2.5, 0.5 );
   exp6.addCoord( 1.5, 0.5 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   exp.addPolygon( exp3 );
   exp.addPolygon( exp4 );
   exp.addPolygon( exp5 );
   exp.addPolygon( exp6 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorPolyFourSides complete");
}

/**
 * Test the xor of two complex, non-convex, non-self-intersecting
 * polygons - V overlap
 */
void testXorPolyVOverlaps()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 1.0, 0.0 );
   p1.addCoord( 5.0, 0.0 );
   p1.addCoord( 5.0, 5.0 );
   p1.addCoord( 3.0, 3.0 );
   p1.addCoord( 1.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon();
   p2.addCoord( 0.0, 2.0 );
   p2.addCoord( 2.0, 0.0 );
   p2.addCoord( 3.0, 1.0 );
   p2.addCoord( 4.0, 0.0 );
   p2.addCoord( 6.0, 2.0 );
   p2.addCoord( 4.0, 4.0 );
   p2.addCoord( 3.0, 3.0 );
   p2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp1 = new Geo.ComplexPolygon();
   exp1.addCoord( 5.0, 5.0 );
   exp1.addCoord( 4.0, 4.0 );
   exp1.addCoord( 5.0, 3.0 );
   
   Geo.Polygon exp2 = new Geo.ComplexPolygon();
   exp2.addCoord( 1.0, 5.0 );
   exp2.addCoord( 1.0, 3.0 );
   exp2.addCoord( 2.0, 4.0 );
   
   Geo.Polygon exp3 = new Geo.ComplexPolygon();
   exp3.addCoord( 5.0, 3.0 );
   exp3.addCoord( 5.0, 1.0 );
   exp3.addCoord( 6.0, 2.0 );
   
   Geo.Polygon exp4 = new Geo.ComplexPolygon();
   exp4.addCoord( 1.0, 3.0 );
   exp4.addCoord( 0.0, 2.0 );
   exp4.addCoord( 1.0, 1.0 );
   
   Geo.Polygon exp5 = new Geo.ComplexPolygon();
   exp5.addCoord( 5.0, 1.0 );
   exp5.addCoord( 4.0, 0.0 );
   exp5.addCoord( 5.0, 0.0 );
   
   Geo.Polygon exp6 = new Geo.ComplexPolygon();
   exp6.addCoord( 3.0, 1.0 );
   exp6.addCoord( 2.0, 0.0 );
   exp6.addCoord( 4.0, 0.0 );
   
   Geo.Polygon exp7 = new Geo.ComplexPolygon();
   exp7.addCoord( 1.0, 1.0 );
   exp7.addCoord( 1.0, 0.0 );
   exp7.addCoord( 2.0, 0.0 );
   
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addPolygon( exp1 );
   exp.addPolygon( exp2 );
   exp.addPolygon( exp3 );
   exp.addPolygon( exp4 );
   exp.addPolygon( exp5 );
   exp.addPolygon( exp6 );
   exp.addPolygon( exp7 );
   
   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p1, p2 );
   //result.print();
   assert( exp == result );
   print("testXorPolyVOverlaps complete");
}
   
/**
 * Test the xor of a rectangle with a hole and solid rectangle
 */
void testXorRectangleHole()
{
   Geo.Polygon p1 = new Geo.ComplexPolygon();
   p1.addCoord( 0.0, 1.0 );
   p1.addCoord( 4.0, 1.0 );
   p1.addCoord( 4.0, 5.0 );
   p1.addCoord( 0.0, 5.0 );
   
   Geo.Polygon p2 = new Geo.ComplexPolygon.asHole(true);
   p2.addCoord( 1.0, 2.0 );
   p2.addCoord( 3.0, 2.0 );
   p2.addCoord( 3.0, 4.0 );
   p2.addCoord( 1.0, 4.0 );
   
   Geo.Polygon p12 = new Geo.ComplexPolygon();
   p12.addPolygon( p1 );
   p12.addPolygon( p2 );

   Geo.Polygon p3 = new Geo.ComplexPolygon();
   p3.addCoord( 2.0, 0.0 );
   p3.addCoord( 6.0, 0.0 );
   p3.addCoord( 6.0, 6.0 );
   p3.addCoord( 2.0, 6.0 );

   // -----------------------------------------------------------
   // --- This is not what I expected and it seems reasonable ---
   // --- However it could be wrong.                          ---
   // -----------------------------------------------------------
   // --- I computed the area of this poly and it came out to ---
   // --- be 24 which is what you would expect.               ---
   // -----------------------------------------------------------
   Geo.Polygon exp = new Geo.ComplexPolygon();
   exp.addCoord( 6.0, 6.0 );
   exp.addCoord( 2.0, 6.0 );
   exp.addCoord( 2.0, 5.0 );
   exp.addCoord( 4.0, 5.0 );
   exp.addCoord( 4.0, 1.0 );
   exp.addCoord( 2.0, 1.0 );
   exp.addCoord( 2.0, 2.0 );
   exp.addCoord( 1.0, 2.0 );
   exp.addCoord( 1.0, 4.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 2.0, 2.0 );
   exp.addCoord( 3.0, 2.0 );
   exp.addCoord( 3.0, 4.0 );
   exp.addCoord( 2.0, 4.0 );
   exp.addCoord( 2.0, 5.0 );
   exp.addCoord( 0.0, 5.0 );
   exp.addCoord( 0.0, 1.0 );
   exp.addCoord( 2.0, 1.0 );
   exp.addCoord( 2.0, 0.0 );
   exp.addCoord( 6.0, 0.0 );

   Geo.ComplexPolygon result = Geo.Boolean.xorComplex( p12, p3 );
   //result.print();
   assert( exp == result );
   print("testXorRectangleHole complete");
}

/**
 * Test the equality operator when empty
 */
void testEqualsEmpty() {
  Geo.Polygon p0a = new Geo.ComplexPolygon();
  Geo.Polygon p0b = new Geo.ComplexPolygon();

    //assert(!(p0a == p0b));
    assert(p0a == p0b);
    assert(p0b == p0a);
    assert(p0a == p0a);
    print("testEqualsEmpty complete");
}

/**
 * Test the equality operator when same
 */
void testEqualsSame() {
  Geo.Polygon p1a = new Geo.ComplexPolygon();
    p1a.addCoord(0.0, 0.0);
    p1a.addCoord(4.0, 0.0);
    p1a.addCoord(4.0, 4.0);
    p1a.addCoord(0.0, 4.0);

    Geo.Polygon p1b = new Geo.ComplexPolygon();
    p1b.addCoord(0.0, 0.0);
    p1b.addCoord(4.0, 0.0);
    p1b.addCoord(4.0, 4.0);
    p1b.addCoord(0.0, 4.0);

    //assert(!(p1a == p1b));
    assert(p1a == p1b);
    assert(p1b == p1a);
    assert(p1a == p1a);
    print("testEqualsSame complete");
}

/**
 * Test the equality operator when same but different order
 */
void testEqualsDifferentOrder() {
  Geo.Polygon p1a = new Geo.ComplexPolygon();
    p1a.addCoord(0.0, 0.0);
    p1a.addCoord(4.0, 0.0);
    p1a.addCoord(4.0, 4.0);
    p1a.addCoord(0.0, 4.0);

    Geo.Polygon p1b = new Geo.ComplexPolygon();
    p1b.addCoord(0.0, 4.0);
    p1b.addCoord(0.0, 0.0);
    p1b.addCoord(4.0, 0.0);
    p1b.addCoord(4.0, 4.0);

    //assert(!(p1a == p1b));
    assert(p1a == p1b);
    assert(p1b == p1a);
    print("testEqualsDifferentOrder complete");
}

/**
 * Test the equality operator when same points but not it same order
 */
void testEqualsBadOrder() {
  Geo.Polygon p1a = new Geo.ComplexPolygon();
    p1a.addCoord(0.0, 0.0);
    p1a.addCoord(4.0, 0.0);
    p1a.addCoord(4.0, 4.0);
    p1a.addCoord(0.0, 4.0);

    Geo.Polygon p1b = new Geo.ComplexPolygon();
    p1b.addCoord(0.0, 0.0);
    p1b.addCoord(0.0, 4.0);
    p1b.addCoord(4.0, 0.0);
    p1b.addCoord(4.0, 4.0);

    assert(!(p1a == p1b));
    assert(!(p1b == p1a));
    print("testEqualsBadOrder complete");
}

/**
 * Test the equality operator when polys with different number of points
 */
void testEqualsDifferentNumPoints() {
  Geo.Polygon p1a = new Geo.ComplexPolygon();
    p1a.addCoord(0.0, 0.0);
    p1a.addCoord(4.0, 0.0);

    Geo.Polygon p1b = new Geo.ComplexPolygon();
    p1b.addCoord(0.0, 0.0);
    p1b.addCoord(4.0, 0.0);
    p1b.addCoord(4.0, 4.0);
    p1b.addCoord(0.0, 4.0);

    assert(!(p1a == p1b));
    assert(!(p1b == p1a));
    print("testEqualsDifferentNumPoints complete");
}

/**
 * Test the equality operator when smae number of points but one point value
 * different
 */
void testEqualsDifferentValue() {
  Geo.Polygon p1a = new Geo.ComplexPolygon();
    p1a.addCoord(0.0, 0.0);
    p1a.addCoord(4.0, 0.0);
    p1a.addCoord(4.0, 4.0);
    p1a.addCoord(0.0, 4.0);

    Geo.Polygon p1b = new Geo.ComplexPolygon();
    p1b.addCoord(0.0, 0.0);
    p1b.addCoord(4.0, 0.0);
    p1b.addCoord(4.0, 4444444.0);
    p1b.addCoord(0.0, 4.0);

    assert(!(p1a == p1b));
    assert(!(p1b == p1a));
    print("testEqualsDifferentValue complete");
}

/**
 * Test the getting the area of a triangle
 */
void testAreaTriangle() {
  Geo.Polygon p = new Geo.ComplexPolygon();
    p.addCoord(0.0, 0.0);
    p.addCoord(2.0, 0.0);
    p.addCoord(2.0, 2.0);

    assert(2.0 == p.area);
    print("testAreaTriangle complete");
}

/**
 * Test the getting the area of a square
 */
void testAreaSquare() {
  Geo.Polygon p = new Geo.ComplexPolygon();
    p.addCoord(0.0, 0.0);
    p.addCoord(2.0, 0.0);
    p.addCoord(2.0, 2.0);
    p.addCoord(0.0, 2.0);

    assert(4.0 == p.area);
    print("testAreaSquare complete");
}

/**
 * Test the getting the area of a non-convex polygon
 */
void testAreaNonConvex() {
  Geo.Polygon p = new Geo.ComplexPolygon();
    p.addCoord(0.0, 0.0);
    p.addCoord(0.0, 8.0);
    p.addCoord(6.0, 4.0);
    p.addCoord(4.0, 2.0);
    p.addCoord(2.0, 4.0);

    assert(20.0 == p.area);
    print("testAreaNonConvex complete");
}

/**
 * Test the getting the area of a polygon with a hole
 */
void testAreaWithHole() {
  Geo.Polygon p1 = new Geo.ComplexPolygon();
    p1.addCoord(0.0, 0.0);
    p1.addCoord(4.0, 0.0);
    p1.addCoord(4.0, 4.0);
    p1.addCoord(0.0, 4.0);

    Geo.Polygon p2 = new Geo.ComplexPolygon.asHole(true);
    p2.addCoord(1.0, 1.0);
    p2.addCoord(3.0, 1.0);
    p2.addCoord(3.0, 3.0);
    p2.addCoord(1.0, 3.0);

    Geo.Polygon p = new Geo.ComplexPolygon();
    p.addPolygon(p1);
    p.addPolygon(p2);

    double area = p.area;
    assert(12.0 == area);
    print("testAreaWithHole complete");
}

/**
 * Test the getting the area of a self-intersecting polygon
 */
void testSelfIntersecting() {
  Geo.Polygon p = new Geo.ComplexPolygon();
    p.addCoord(6.0, 6.0);
    p.addCoord(2.0, 6.0);
    p.addCoord(2.0, 5.0);
    p.addCoord(4.0, 5.0);
    p.addCoord(4.0, 1.0);
    p.addCoord(2.0, 1.0);
    p.addCoord(2.0, 2.0);
    p.addCoord(1.0, 2.0);
    p.addCoord(1.0, 4.0);
    p.addCoord(2.0, 4.0);
    p.addCoord(2.0, 2.0);
    p.addCoord(3.0, 2.0);
    p.addCoord(3.0, 4.0);
    p.addCoord(2.0, 4.0);
    p.addCoord(2.0, 5.0);
    p.addCoord(0.0, 5.0);
    p.addCoord(0.0, 1.0);
    p.addCoord(2.0, 1.0);
    p.addCoord(2.0, 0.0);
    p.addCoord(6.0, 0.0);

    assert(24.0 == p.area);
    print("testSelfIntersecting complete");
}
